import collections
import numpy as np
import os
import packaging.version
import pandas as pd
import skimage.io
import skimage.measure
import tempfile
import histomicstk.preprocessing.color_normalization as htk_cnorm
import histomicstk.preprocessing.color_deconvolution as htk_cdeconv
import histomicstk.segmentation.nuclear as htk_nuclear
import histomicstk.features as htk_features
import sys
thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, thisDir)
import htk_test_utilities as utilities  # noqa

try:
    from .datastore import datastore  # noqa
except ModuleNotFoundError:  # local prototyping
    from datastore import datastore  # noqa


class Cfg:
    def __init__(self):
        self.fdata_nuclei = None
        self.im_nuclei_stain = None
        self.im_nuclei_seg_mask = None
        self.nuclei_rprops = None


cfg = Cfg()
# Enable to generate groundtruth files in the /tmp directory
GENERATE_GROUNDTRUTH = False


def check_fdata_sanity(fdata, expected_feature_list,
                       prefix='', match_feature_count=True):

    assert len(cfg.nuclei_rprops) == fdata.shape[0]

    if len(prefix) > 0:
        fcols = [col for col in fdata.columns if col.startswith(prefix)]
    else:
        fcols = fdata.columns

    if match_feature_count:
        assert len(fcols) == len(expected_feature_list)

    for col in expected_feature_list:
        assert prefix + col in fcols


class TestFeatureExtraction:

    def test_setup(self):

        # define parameters
        args = {
            'reference_mu_lab': [8.63234435, -0.11501964, 0.03868433],
            'reference_std_lab': [0.57506023, 0.10403329, 0.01364062],

            'min_radius': 12,
            'max_radius': 30,
            'foreground_threshold': 60,
            'min_nucleus_area': 80,
            'local_max_search_radius': 10,
        }

        args = collections.namedtuple('Parameters', args.keys())(**args)

        # read input image
        input_image_file = datastore.fetch('Easy1.png')

        im_input = skimage.io.imread(input_image_file)[:, :, :3]

        # perform color normalization
        im_input_nmzd = htk_cnorm.reinhard(
            im_input, args.reference_mu_lab, args.reference_std_lab)

        # perform color decovolution
        w = htk_cdeconv.rgb_separate_stains_macenko_pca(
            im_input_nmzd, im_input_nmzd.max())

        im_stains = htk_cdeconv.color_deconvolution(im_input_nmzd, w).Stains

        nuclei_channel = htk_cdeconv.find_stain_index(
            htk_cdeconv.stain_color_map['hematoxylin'], w)

        im_nuclei_stain = im_stains[:, :, nuclei_channel].astype(np.float)

        cytoplasm_channel = htk_cdeconv.find_stain_index(
            htk_cdeconv.stain_color_map['eosin'], w)

        im_cytoplasm_stain = im_stains[:, :, cytoplasm_channel].astype(
            np.float)

        # segment nuclei
        im_nuclei_seg_mask = htk_nuclear.detect_nuclei_kofahi(
            im_nuclei_stain,
            im_nuclei_stain < args.foreground_threshold,
            args.min_radius,
            args.max_radius,
            args.min_nucleus_area,
            args.local_max_search_radius
        )

        # perform connected component analysis
        nuclei_rprops = skimage.measure.regionprops(im_nuclei_seg_mask)

        # compute nuclei features
        fdata_nuclei = htk_features.compute_nuclei_features(
            im_nuclei_seg_mask, im_nuclei_stain,
            im_cytoplasm=im_cytoplasm_stain)

        cfg.im_input = im_input
        cfg.im_input_nmzd = im_input_nmzd
        cfg.im_nuclei_stain = im_nuclei_stain
        cfg.im_nuclei_seg_mask = im_nuclei_seg_mask
        cfg.nuclei_rprops = nuclei_rprops
        cfg.fdata_nuclei = fdata_nuclei

    def test_compute_intensity_features(self):

        expected_feature_list = [
            'Intensity.Min',
            'Intensity.Max',
            'Intensity.Mean',
            'Intensity.Median',
            'Intensity.MeanMedianDiff',
            'Intensity.Std',
            'Intensity.IQR',
            'Intensity.MAD',
            'Intensity.Skewness',
            'Intensity.Kurtosis',
            'Intensity.HistEnergy',
            'Intensity.HistEntropy',
        ]

        fdata = htk_features.compute_intensity_features(
            cfg.im_nuclei_seg_mask, cfg.im_nuclei_stain)

        check_fdata_sanity(fdata, expected_feature_list)

        check_fdata_sanity(
            cfg.fdata_nuclei, expected_feature_list,
            prefix='Nucleus.', match_feature_count=False)

        check_fdata_sanity(
            cfg.fdata_nuclei, expected_feature_list,
            prefix='Cytoplasm.', match_feature_count=False)

        if GENERATE_GROUNDTRUTH:
            fdata.to_csv(os.path.join(
                tempfile.gettempdir(), 'Easy1_nuclei_intensity_features.csv'),
                index=False)

        fdata_gtruth = pd.read_csv(
            utilities.getTestFilePath('Easy1_nuclei_intensity_features.csv'),
            index_col=None)

        pd.testing.assert_frame_equal(
            fdata, fdata_gtruth, check_less_precise=2)

    def test_compute_haralick_features(self):

        f = [
            'Haralick.ASM',
            'Haralick.Contrast',
            'Haralick.Correlation',
            'Haralick.SumOfSquares',
            'Haralick.IDM',
            'Haralick.SumAverage',
            'Haralick.SumVariance',
            'Haralick.SumEntropy',
            'Haralick.Entropy',
            'Haralick.DifferenceVariance',
            'Haralick.DifferenceEntropy',
            'Haralick.IMC1',
            'Haralick.IMC2',
        ]

        expected_feature_list = []
        for col in f:
            expected_feature_list.append(col + '.Mean')
            expected_feature_list.append(col + '.Range')

        fdata = htk_features.compute_haralick_features(
            cfg.im_nuclei_seg_mask, cfg.im_nuclei_stain)

        check_fdata_sanity(fdata, expected_feature_list)

        check_fdata_sanity(
            cfg.fdata_nuclei, expected_feature_list,
            prefix='Nucleus.', match_feature_count=False)

        check_fdata_sanity(
            cfg.fdata_nuclei, expected_feature_list,
            prefix='Cytoplasm.', match_feature_count=False)

        if GENERATE_GROUNDTRUTH:
            fdata.to_csv(os.path.join(
                tempfile.gettempdir(), 'Easy1_nuclei_haralick_features.csv'),
                index=False)

        fdata_gtruth = pd.read_csv(
            utilities.getTestFilePath('Easy1_nuclei_haralick_features.csv'))

        pd.testing.assert_frame_equal(
            fdata, fdata_gtruth, check_less_precise=2)

    def test_compute_gradient_features(self):

        expected_feature_list = [
            'Gradient.Mag.Mean',
            'Gradient.Mag.Std',
            'Gradient.Mag.Skewness',
            'Gradient.Mag.Kurtosis',
            'Gradient.Mag.HistEntropy',
            'Gradient.Mag.HistEnergy',
            'Gradient.Canny.Sum',
            'Gradient.Canny.Mean',
        ]

        fdata = htk_features.compute_gradient_features(
            cfg.im_nuclei_seg_mask, cfg.im_nuclei_stain)

        check_fdata_sanity(fdata, expected_feature_list)

        check_fdata_sanity(
            cfg.fdata_nuclei, expected_feature_list,
            prefix='Nucleus.', match_feature_count=False)

        check_fdata_sanity(
            cfg.fdata_nuclei, expected_feature_list,
            prefix='Cytoplasm.', match_feature_count=False)

        if GENERATE_GROUNDTRUTH:
            fdata.to_csv(os.path.join(
                tempfile.gettempdir(), 'Easy1_nuclei_gradient_features.csv'),
                index=False)

        fdata_gtruth = pd.read_csv(
            utilities.getTestFilePath('Easy1_nuclei_gradient_features.csv'),
            index_col=None)

        pd.testing.assert_frame_equal(
            fdata, fdata_gtruth, check_less_precise=2)

    def test_compute_morphometry_features(self):

        expected_feature_list = [
            'Orientation.Orientation',
            'Size.Area',
            'Size.ConvexHullArea',
            'Size.MajorAxisLength',
            'Size.MinorAxisLength',
            'Size.Perimeter',
            'Shape.Circularity',
            'Shape.Eccentricity',
            'Shape.EquivalentDiameter',
            'Shape.Extent',
            'Shape.FractalDimension',
            'Shape.MinorMajorAxisRatio',
            'Shape.Solidity',
            'Shape.HuMoments1',
            'Shape.HuMoments2',
            'Shape.HuMoments3',
            'Shape.HuMoments4',
            'Shape.HuMoments5',
            'Shape.HuMoments6',
            'Shape.HuMoments7',
        ]

        fdata = htk_features.compute_morphometry_features(
            cfg.im_nuclei_seg_mask)

        check_fdata_sanity(fdata, expected_feature_list)

        check_fdata_sanity(
            cfg.fdata_nuclei, expected_feature_list, match_feature_count=False)

        if GENERATE_GROUNDTRUTH:
            fdata.to_csv(os.path.join(
                tempfile.gettempdir(),
                'Easy1_nuclei_morphometry_features.csv'),
                index=False)

        if packaging.version.parse(skimage.__version__) < packaging.version.parse('0.18'):
            fdata_gtruth = pd.read_csv(
                utilities.getTestFilePath('Easy1_nuclei_morphometry_features.csv'),
                index_col=None)
            pd.testing.assert_frame_equal(
                fdata, fdata_gtruth, check_less_precise=2)

    def test_compute_fsd_features(self):

        Fs = 6
        expected_feature_list = ['Shape.FSD' + str(i + 1) for i in range(Fs)]

        fdata = htk_features.compute_fsd_features(
            cfg.im_nuclei_seg_mask, Fs=Fs)

        check_fdata_sanity(fdata, expected_feature_list)

        check_fdata_sanity(
            cfg.fdata_nuclei, expected_feature_list, match_feature_count=False)

        if GENERATE_GROUNDTRUTH:
            fdata.to_csv(os.path.join(
                tempfile.gettempdir(), 'Easy1_nuclei_fsd_features.csv'),
                index=False)

        fdata_gtruth = pd.read_csv(
            utilities.getTestFilePath('Easy1_nuclei_fsd_features.csv'))

        pd.testing.assert_frame_equal(
            fdata, fdata_gtruth, check_less_precise=2)


# PROTOTYPING =================================================================

if __name__ == '__main__':

    tfe = TestFeatureExtraction()
    tfe.test_setup()
    tfe.test_compute_morphometry_features()
    tfe.test_compute_intensity_features()



#!/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import os
import numpy as np
import skimage.io
import histomicstk.preprocessing.color_conversion as htk_cvt
import histomicstk.preprocessing.color_normalization as htk_cnorm
import histomicstk.preprocessing.color_deconvolution as htk_cdeconv
import histomicstk.segmentation as htk_seg

from .datastore import datastore


class TestNucleiSegmentation:

    def test_segment_nuclei_kofahi(self):

        input_image_file = datastore.fetch('Easy1.png')

        ref_image_file = datastore.fetch('L1.png')

        # read input image
        im_input = skimage.io.imread(input_image_file)[:, :, :3]

        # read reference image
        im_reference = skimage.io.imread(ref_image_file)[:, :, :3]

        # get mean and stddev of reference image in lab space
        mean_ref, std_ref = htk_cvt.lab_mean_std(im_reference)

        # perform color normalization
        im_nmzd = htk_cnorm.reinhard(im_input, mean_ref, std_ref)

        # perform color decovolution
        stain_color_map = {
            'hematoxylin': [0.65, 0.70, 0.29],
            'eosin': [0.07, 0.99, 0.11],
            'dab': [0.27, 0.57, 0.78],
            'null': [0.0, 0.0, 0.0]
        }

        w = htk_cdeconv.rgb_separate_stains_macenko_pca(im_nmzd, im_nmzd.max())

        im_stains = htk_cdeconv.color_deconvolution(im_nmzd, w).Stains

        nuclei_channel = htk_cdeconv.find_stain_index(stain_color_map['hematoxylin'], w)

        im_nuclei_stain = im_stains[:, :, nuclei_channel].astype(np.float)

        # segment nuclei
        im_nuclei_seg_mask = htk_seg.nuclear.detect_nuclei_kofahi(
            im_nuclei_stain, im_nuclei_stain < 60,
            min_radius=20, max_radius=30,
            min_nucleus_area=80, local_max_search_radius=10
        )

        num_nuclei = len(np.unique(im_nuclei_seg_mask)) - 1

        # check if segmentation mask matches ground truth
        gtruth_mask_file = os.path.join(datastore.fetch(
            'Easy1_nuclei_seg_kofahi.npy'))

        im_gtruth_mask = np.load(gtruth_mask_file)

        num_nuclei_gtruth = len(np.unique(im_gtruth_mask)) - 1

        assert num_nuclei == num_nuclei_gtruth

        np.testing.assert_allclose(im_nuclei_seg_mask, im_gtruth_mask)

        # check no nuclei case
        im_nuclei_seg_mask = htk_seg.nuclear.detect_nuclei_kofahi(
            255 * np.ones_like(im_nuclei_stain), np.ones_like(im_nuclei_stain),
            min_radius=20, max_radius=30,
            min_nucleus_area=80, local_max_search_radius=10
        )

        num_nuclei = len(np.unique(im_nuclei_seg_mask)) - 1

        assert num_nuclei == 0



#!/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import collections
import numpy as np
import os
from histomicstk.preprocessing import color_normalization as htk_cn
from histomicstk.cli import utils as cli_utils

from .datastore import datastore


class TestReinhardNormalization:

    def test_reinhard_stats(self):

        wsi_path = os.path.join(datastore.fetch(
            'sample_svs_image.TCGA-DU-6399-01A-01-TS1.e8eb65de-d63e-42db-af6f-14fefbbdf7bd.svs'  # noqa
        ))

        np.random.seed(1)

        # create dask client
        args = {
            # In Python 3 unittesting, the scheduler fails if it uses processes
            'scheduler': 'multithreading',  # None,
            'num_workers': -1,
            'num_threads_per_worker': 1,
        }

        args = collections.namedtuple('Parameters', args.keys())(**args)

        cli_utils.create_dask_client(args)

        # compute reinhard stats
        wsi_mean, wsi_stddev = htk_cn.reinhard_stats(
            wsi_path, 0.1, magnification=20)

        gt_mean = [8.896134, -0.074579, 0.022006]
        gt_stddev = [0.612143, 0.122667, 0.021361]

        np.testing.assert_allclose(wsi_mean, gt_mean, atol=1e-2)
        np.testing.assert_allclose(wsi_stddev, gt_stddev, atol=1e-2)


class TestBackgroundIntensity:

    def test_background_intensity(self):
        wsi_path = os.path.join(datastore.fetch(
            'sample_svs_image.TCGA-DU-6399-01A-01-TS1.e8eb65de-d63e-42db-af6f-14fefbbdf7bd.svs'
        ))

        np.random.seed(1)

        # create dask client
        args = {
            # In Python 3 unittesting, the scheduler fails if it uses processes
            'scheduler': 'multithreading',  # None,
            'num_workers': -1,
            'num_threads_per_worker': 1,
        }

        args = collections.namedtuple('Parameters', args.keys())(**args)

        cli_utils.create_dask_client(args)

        # compute background intensity
        I_0 = htk_cn.background_intensity(wsi_path,
                                          sample_approximate_total=5000)

        np.testing.assert_allclose(I_0, [242, 244, 241], atol=1)



"""
This package contains implementations of state-of-th-art methods for
segmenting nuclei from histopathology images.
"""

from .gaussian_voting import gaussian_voting
from .gvf_tracking import gvf_tracking
from .max_clustering import max_clustering
from .min_model import min_model
from .detect_nuclei_kofahi import detect_nuclei_kofahi

__all__ = (
    'detect_nuclei_kofahi',
    'gaussian_voting',
    'gvf_tracking',
    'max_clustering',
    'min_model'
)



# import sub-packages to support nested calls
from . import utils  # must be imported before other packages

from . import segmentation  # must be imported before features

from . import features
from . import filters
from . import preprocessing
from . import annotations_and_masks
from . import saliency
from . import workflows

# list out things that are available for public use
__all__ = (

    # sub-packages
    'features',
    'filters',
    'preprocessing',
    'segmentation',
    'utils',
    'annotations_and_masks',
    'saliency',
    'workflows',
)



import skimage.io

from histomicstk.cli.utils import CLIArgumentParser
import large_image

import histomicstk.preprocessing.color_deconvolution as htk_cd

from histomicstk.cli import utils

import logging
logging.basicConfig()


def main(args):

    # Read Input Image
    print('>> Reading input image')

    print(args.inputImageFile)

    ts = large_image.getTileSource(args.inputImageFile)

    im_input = ts.getRegion(
        format=large_image.tilesource.TILE_FORMAT_NUMPY,
        **utils.get_region_dict(args.region, args.maxRegionSize, ts)
    )[0]

    # Create stain matrix
    print('>> Creating stain matrix')

    w = utils.get_stain_matrix(args)
    print(w)

    # Perform color deconvolution
    print('>> Performing color deconvolution')
    im_stains = htk_cd.color_deconvolution(im_input, w).Stains

    # write stain images to output
    print('>> Outputting individual stain images')

    print(args.outputStainImageFile_1)
    skimage.io.imsave(args.outputStainImageFile_1, im_stains[:, :, 0])

    print(args.outputStainImageFile_2)
    skimage.io.imsave(args.outputStainImageFile_2, im_stains[:, :, 1])

    print(args.outputStainImageFile_3)
    skimage.io.imsave(args.outputStainImageFile_3, im_stains[:, :, 2])


if __name__ == "__main__":
    main(CLIArgumentParser().parse_args())



import cherrypy
import json
import os

from girder.models.assetstore import Assetstore
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.user import User
from girder.utility.progress import ProgressContext
from girder.utility.server import configureServer

from girder_large_image_annotation.models.annotation import Annotation

from datastore import datastore


def namedFolder(user, folderName='Public'):
    return Folder().find({
        'parentId': user['_id'],
        'name': folderName,
    })[0]


cherrypy.config['database']['uri'] = 'mongodb://mongodb:27017/girder'

# This loads plugins, allowing setting validation
configureServer()

if User().findOne() is None:
    User().createUser('admin', 'password', 'Admin', 'Admin', 'admin@nowhere.nil')
adminUser = User().findOne({'admin': True})
if Assetstore().findOne() is None:
    Assetstore().createFilesystemAssetstore('Assetstore', '/assetstore')
fsAssetstore = Assetstore().findOne()
# Upload a list of items
publicFolder = namedFolder(adminUser)
dataFiles = {
    'tcga1': {
        'name': 'TCGA-A2-A0YE-01Z-00-DX1.8A2E3094-5755-42BC-969D-7F0A2ECA0F39.svs'},
}
for key, entry in dataFiles.items():
    path = datastore.fetch(entry['name'])
    query = {'folderId': publicFolder['_id'], 'name': os.path.basename(path)}
    if not Item().findOne(query):
        with ProgressContext(False, user=adminUser) as ctx:
            Assetstore().importData(
                fsAssetstore, publicFolder, 'folder', {'importPath': path},
                ctx, adminUser, leafFoldersAsItems=False)
    entry['item'] = Item().findOne(query)
    print(entry['item'])
# Add annotations
annotationFiles = {
    'tcga1': {
        'item': 'tcga1',
        'name': 'TCGA-A2-A0YE-01Z-00-DX1.8A2E3094-5755-42BC-969D-'
                '7F0A2ECA0F39.svs_annotations.json',
    }
}
for key, entry in annotationFiles.items():
    item = dataFiles[entry['item']]['item']
    query = {'_active': {'$ne': False}, 'itemId': item['_id']}
    if not Annotation().findOne(query):
        path = datastore.fetch(entry['name'])
        annotations = json.load(open(path))
        if not isinstance(annotations, list):
            annotations = [annotations]
        for annotation in annotations:
            Annotation().createAnnotation(
                item, adminUser, annotation.get('annotation', annotation))
    print(Annotation().find(query).count())



import os

from histomicstk.cli.utils import CLIArgumentParser
import large_image
import numpy as np
import skimage.io

import histomicstk.segmentation.positive_pixel_count as ppc

from histomicstk.cli import utils


def main(args):
    utils.create_dask_client(args)
    ts = large_image.getTileSource(args.inputImageFile)
    make_label_image = getattr(args, 'outputLabelImage', None) is not None
    region = utils.get_region_dict(
        args.region,
        *(args.maxRegionSize, ts) if make_label_image else ()
    ).get('region')
    ppc_params = ppc.Parameters(
        **{k: getattr(args, k) for k in ppc.Parameters._fields}
    )
    results = ppc.count_slide(
        args.inputImageFile, ppc_params, region,
        args.tile_grouping, make_label_image,
    )
    if make_label_image:
        stats, label_image = results
        # Colorize label image.  Colors from the "coolwarm" color map
        color_map = np.empty((4, 3), dtype=np.uint8)
        color_map[ppc.Labels.NEGATIVE] = 255
        color_map[ppc.Labels.WEAK] = 60, 78, 194
        color_map[ppc.Labels.PLAIN] = 221, 220, 220
        color_map[ppc.Labels.STRONG] = 180, 4, 38
        # Cleverly index color_map
        label_image = color_map[label_image]
        try:
            skimage.io.imsave(args.outputLabelImage, label_image)
        except ValueError:
            # This is likely caused by an unknown extension, so try again
            altname = args.outputLabelImage + '.png'
            skimage.io.imsave(altname, label_image)
            os.rename(altname, args.outputLabelImage)
    else:
        stats, = results
    with open(args.returnParameterFile, 'w') as f:
        for k, v in zip(stats._fields, stats):
            f.write(f'{k} = {v}\n')


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



#!/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

# This is to serve as an example for how to create a server-side test in a
# girder plugin, it is not meant to be useful.

from argparse import Namespace
import os
import json
import collections
import numpy as np
import skimage.io
import large_image
import histomicstk.preprocessing.color_deconvolution as htk_cdeconv
import histomicstk.preprocessing.color_normalization as htk_cnorm
import histomicstk.segmentation.nuclear as htk_nuclear
import histomicstk.utils as htk_utils
from histomicstk.cli import utils as cli_utils

from .datastore import datastore


class TestCliCommon:

    def test_get_stain_matrix(self):

        args = Namespace(
            stain_1='hematoxylin',
            stain_1_vector=[-1, -1, -1],
            stain_2='custom',
            stain_2_vector=[0.1, 0.2, 0.3],
        )
        expected = np.array([htk_cdeconv.stain_color_map['hematoxylin'],
                             [0.1, 0.2, 0.3]]).T
        np.testing.assert_allclose(cli_utils.get_stain_matrix(args, 2),
                                   expected)

    def test_get_region_dict(self):

        ts = large_image.getTileSource(os.path.join(
            datastore.fetch('Easy1.png')))

        result = cli_utils.get_region_dict([-1, -1, -1, -1], 2000, ts)
        expected = {}
        assert result == expected, "Expected {}, got {}".format(expected,
                                                                result)

        result = cli_utils.get_region_dict([100, 110, 250, 240], 500, ts)
        expected = dict(region=dict(left=100, top=110, width=250, height=240))
        assert result == expected, "Expected {}, got {}".format(expected,
                                                                result)

    def test_segment_wsi_foreground_at_low_res(self):
        np.random.seed(0)

        wsi_path = os.path.join(datastore.fetch(
            'TCGA-06-0129-01Z-00-DX3.bae772ea-dd36-47ec-8185-761989be3cc8.svs'  # noqa
        ))

        ts = large_image.getTileSource(wsi_path)

        im_fgnd_mask_lres, fgnd_seg_scale = \
            cli_utils.segment_wsi_foreground_at_low_res(ts)

        np.testing.assert_equal(fgnd_seg_scale['magnification'], 2.5)

        fgnd_mask_gtruth_file = os.path.join(datastore.fetch(
            'TCGA-06-0129-01Z-00-DX3_fgnd_mask_lres.png'))

        im_fgnd_mask_lres_gtruth = skimage.io.imread(
            fgnd_mask_gtruth_file) > 0

        np.testing.assert_array_equal(im_fgnd_mask_lres > 0,
                                      im_fgnd_mask_lres_gtruth)

    def test_create_tile_nuclei_annotations(self):

        wsi_path = os.path.join(datastore.fetch(
            'TCGA-06-0129-01Z-00-DX3.bae772ea-dd36-47ec-8185-761989be3cc8.svs'  # noqa
        ))

        # define parameters
        args = {

            'reference_mu_lab': [8.63234435, -0.11501964, 0.03868433],
            'reference_std_lab': [0.57506023, 0.10403329, 0.01364062],

            'stain_1': 'hematoxylin',
            'stain_2': 'eosin',
            'stain_3': 'null',

            'stain_1_vector': [-1, -1, -1],
            'stain_2_vector': [-1, -1, -1],
            'stain_3_vector': [-1, -1, -1],

            'min_fgnd_frac': 0.50,
            'analysis_mag': 20,
            'analysis_tile_size': 1200,

            'foreground_threshold': 60,
            'min_radius': 6,
            'max_radius': 12,
            'min_nucleus_area': 25,
            'local_max_search_radius': 8,

            # In Python 3 unittesting, the scheduler fails if it uses processes
            'scheduler': 'multithreading',  # None,
            'num_workers': -1,
            'num_threads_per_worker': 1,
        }

        args = collections.namedtuple('Parameters', args.keys())(**args)

        # read WSI
        ts = large_image.getTileSource(wsi_path)

        ts_metadata = ts.getMetadata()

        analysis_tile_size = {
            'width': int(ts_metadata['tileWidth'] * np.floor(
                1.0 * args.analysis_tile_size / ts_metadata['tileWidth'])),
            'height': int(ts_metadata['tileHeight'] * np.floor(
                1.0 * args.analysis_tile_size / ts_metadata['tileHeight']))
        }

        # define ROI
        roi = {'left': ts_metadata['sizeX'] / 2,
               'top': ts_metadata['sizeY'] * 3 / 4,
               'width': analysis_tile_size['width'],
               'height': analysis_tile_size['height'],
               'units': 'base_pixels'}

        # define tile iterator parameters
        it_kwargs = {
            'tile_size': {'width': args.analysis_tile_size},
            'scale': {'magnification': args.analysis_mag},
            'region': roi
        }

        # create dask client
        cli_utils.create_dask_client(args)

        # get tile foregreoung at low res
        im_fgnd_mask_lres, fgnd_seg_scale = \
            cli_utils.segment_wsi_foreground_at_low_res(ts)

        # compute tile foreground fraction
        tile_fgnd_frac_list = htk_utils.compute_tile_foreground_fraction(
            wsi_path, im_fgnd_mask_lres, fgnd_seg_scale,
            it_kwargs
        )

        num_fgnd_tiles = np.count_nonzero(
            tile_fgnd_frac_list >= args.min_fgnd_frac)

        np.testing.assert_equal(num_fgnd_tiles, 2)

        # create nuclei annotations
        nuclei_bbox_annot_list = []
        nuclei_bndry_annot_list = []

        for tile_info in ts.tileIterator(
                format=large_image.tilesource.TILE_FORMAT_NUMPY,
                **it_kwargs):

            im_tile = tile_info['tile'][:, :, :3]

            # perform color normalization
            im_nmzd = htk_cnorm.reinhard(im_tile,
                                         args.reference_mu_lab,
                                         args.reference_std_lab)

            # perform color deconvolution
            w = cli_utils.get_stain_matrix(args)

            im_stains = htk_cdeconv.color_deconvolution(im_nmzd, w).Stains

            im_nuclei_stain = im_stains[:, :, 0].astype(np.float)

            # segment nuclei
            im_nuclei_seg_mask = htk_nuclear.detect_nuclei_kofahi(
                im_nuclei_stain,
                im_nuclei_stain < args.foreground_threshold,
                args.min_radius,
                args.max_radius,
                args.min_nucleus_area,
                args.local_max_search_radius
            )

            # generate nuclei annotations as bboxes
            cur_bbox_annot_list = cli_utils.create_tile_nuclei_annotations(
                im_nuclei_seg_mask, tile_info, 'bbox')

            nuclei_bbox_annot_list.extend(cur_bbox_annot_list)

            # generate nuclei annotations as boundaries
            cur_bndry_annot_list = cli_utils.create_tile_nuclei_annotations(
                im_nuclei_seg_mask, tile_info, 'boundary')

            nuclei_bndry_annot_list.extend(cur_bndry_annot_list)

        # compare nuclei bbox annotations with gtruth
        nuclei_bbox_annot_gtruth_file = os.path.join(datastore.fetch(
            'TCGA-06-0129-01Z-00-DX3_roi_nuclei_bbox.anot'  # noqa
        ))

        with open(nuclei_bbox_annot_gtruth_file) as fbbox_annot:
            nuclei_bbox_annot_list_gtruth = json.load(fbbox_annot)['elements']

        # Check that nuclei_bbox_annot_list is nearly equal to
        # nuclei_bbox_annot_list_gtruth
        assert len(nuclei_bbox_annot_list) == len(nuclei_bbox_annot_list_gtruth)
        for pos in range(len(nuclei_bbox_annot_list)):
            np.testing.assert_array_almost_equal(
                nuclei_bbox_annot_list[pos]['center'],
                nuclei_bbox_annot_list_gtruth[pos]['center'], 0)
            np.testing.assert_almost_equal(
                nuclei_bbox_annot_list[pos]['width'],
                nuclei_bbox_annot_list_gtruth[pos]['width'], 1)
            np.testing.assert_almost_equal(
                nuclei_bbox_annot_list[pos]['height'],
                nuclei_bbox_annot_list_gtruth[pos]['height'], 1)

        # compare nuclei boundary annotations with gtruth
        nuclei_bndry_annot_gtruth_file = os.path.join(datastore.fetch(
            'TCGA-06-0129-01Z-00-DX3_roi_nuclei_boundary.anot'
        ))

        with open(nuclei_bndry_annot_gtruth_file) as fbndry_annot:
            nuclei_bndry_annot_list_gtruth = json.load(
                fbndry_annot)['elements']

        assert len(nuclei_bndry_annot_list) == len(nuclei_bndry_annot_list_gtruth)

        for pos in range(len(nuclei_bndry_annot_list)):

            np.testing.assert_array_almost_equal(
                nuclei_bndry_annot_list[pos]['points'],
                nuclei_bndry_annot_list_gtruth[pos]['points'], 0)

    def test_splitArgs(self):
        args = Namespace(
            a=1,
            b_a=2,
            b_b=3,
            b_c_a=4,
            c_a=5,
        )
        split = cli_utils.splitArgs(args)
        assert split == Namespace(
            a=1,
            b=Namespace(
                a=2,
                b=3,
                c_a=4,
            ),
            c=Namespace(
                a=5,
            ),
        )



#!/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import numpy as np
from pandas import DataFrame
from pandas.testing import assert_frame_equal
from histomicstk.features import compute_global_cell_graph_features as cgcgf


class TestGlobalCellGraphFeatures:
    def testSimple(self):
        data = np.array([[-1, -1], [-1, 1], [1, -1], [1, 1], [-.5, -.5], [.5, .5]])
        actual = cgcgf(data, neighbor_distances=0.7 * np.arange(1, 6), neighbor_counts=(3, 5))
        expected = DataFrame(dict(
            delaunay_area_disorder=0.261203874964,
            delaunay_area_mean=0.666666666667,
            delaunay_area_min_max_ratio=0.5,
            delaunay_area_stddev=0.235702260396,
            delaunay_sides_disorder=0.224914917312,
            delaunay_sides_mean=1.55936204046,
            delaunay_sides_min_max_ratio=0.353553390593,
            delaunay_sides_stddev=0.452497141571,
            density_distance_for_neighbors_0_disorder=0.0959519464365,
            density_distance_for_neighbors_0_mean=1.86037961003,
            density_distance_for_neighbors_0_min_max_ratio=0.790569415042,
            density_distance_for_neighbors_0_stddev=0.197453049082,
            density_distance_for_neighbors_1_disorder=0.113918900724,
            density_distance_for_neighbors_1_mean=2.59272486435,
            density_distance_for_neighbors_1_min_max_ratio=0.75,
            density_distance_for_neighbors_1_stddev=0.333333333333,
            density_neighbors_in_distance_0_disorder=np.nan,
            density_neighbors_in_distance_0_mean=0.0,
            density_neighbors_in_distance_0_min_max_ratio=np.nan,
            density_neighbors_in_distance_0_stddev=0.0,
            density_neighbors_in_distance_1_disorder=0.414213562373,
            density_neighbors_in_distance_1_mean=0.666666666667,
            density_neighbors_in_distance_1_min_max_ratio=0.0,
            density_neighbors_in_distance_1_stddev=0.471404520791,
            density_neighbors_in_distance_2_disorder=0.113918900724,
            density_neighbors_in_distance_2_mean=3.66666666667,
            density_neighbors_in_distance_2_min_max_ratio=0.75,
            density_neighbors_in_distance_2_stddev=0.471404520791,
            density_neighbors_in_distance_3_disorder=0.0981124329991,
            density_neighbors_in_distance_3_mean=4.33333333333,
            density_neighbors_in_distance_3_min_max_ratio=0.8,
            density_neighbors_in_distance_3_stddev=0.471404520791,
            density_neighbors_in_distance_4_disorder=0.0,
            density_neighbors_in_distance_4_mean=5.0,
            density_neighbors_in_distance_4_min_max_ratio=1.0,
            density_neighbors_in_distance_4_stddev=0.0,
            mst_branches_disorder=0.252878170232,
            mst_branches_mean=1.19814095698,
            mst_branches_min_max_ratio=0.4472135955,
            mst_branches_stddev=0.405534520356,
            voronoi_area_disorder=0.0,
            voronoi_area_mean=1.6875,
            voronoi_area_min_max_ratio=1.0,
            voronoi_area_stddev=0.0,
            voronoi_max_dist_disorder=0.0,
            voronoi_max_dist_mean=2.12132034356,
            voronoi_max_dist_min_max_ratio=1.0,
            voronoi_max_dist_stddev=0.0,
            voronoi_peri_disorder=0.0,
            voronoi_peri_mean=5.55368876047,
            voronoi_peri_min_max_ratio=1.0,
            voronoi_peri_stddev=0.0,
        ), index=[0])
        assert_frame_equal(actual, expected, check_like=True)



#!/usr/bin/env python

##############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##############################################################################

import hashlib
import io
import json
import os
import PIL.Image
import runpy
import shutil
import sys
import tempfile

from .datastore import datastore

base = None


def setUpModule():
    # We need to start Girder to test endpoint creation
    base.enabledPlugins.append('HistomicsTK')
    base.startServer()


def tearDownModule():
    base.stopServer()


class TestCliResults:
    def _runTest(self, cli_args=(), cli_kwargs=None, outputs=None,  # noqa
                 contains=None, excludes=None):
        """
        Test a cli by calling runpy.  Ensure that output files match a
        sha256 and stdout contains certain phrases and excludes other
        phrases.  A value in the cli_args or cli_kwargs of
        'tmp_<value>' will create a temporary file and use that
        instead.  The same value in the output list can be used to
        test the existence and contents of that file at the end of the
        process.  Note that stdout capturing has limitations -- only
        output written respecting the current value of sys.stdout is
        captured.

        :param cli_args: a tuple or list of args to format and pass to the cli.
        :param cli_kwargs: keyword arguments to format and pass to the cli.
            Each entry is passed as '--<key>=<value>'.
        :param outputs: a dictionary of expected output files and checks to
            perform on them.  The keys are the file paths.  The values are
            dictionaries that may contain:
                'hash': a sha256 lowercase hexdigest that must match the entire
                    file.
                'contains': a list of phrases that must be present in the first
                    256 kb of the file.
        :param contains: a list of phrases that must be present in the stdout
            output of the cli.
        :param excludes: a list of phrases that must be not present in the
            stdout output of the cli.
        :returns: stdout from the test.

        """
        cli_kwargs = {} if cli_kwargs is None else cli_kwargs
        outputs = {} if outputs is None else outputs
        contains = [] if contains is None else contains
        excludes = [] if excludes is None else excludes

        chunkSize = 256 * 1024

        cwd = os.environ.get('CLI_CWD')
        tmppath = tempfile.mkdtemp()
        try:
            cmd = [arg if not arg.startswith('tmp_') else
                   os.path.join(tmppath, arg) for arg in cli_args]
            cmd += ['--%s=%s' % (
                k, v if not v.startswith('tmp_') else os.path.join(tmppath, v))
                for k, v in cli_kwargs.items()]
            stdout = ''
            try:
                old_sys_argv = sys.argv[:]
                old_stdout, old_stderr = sys.stdout, sys.stderr
                old_cwd = os.getcwd()
                sys.argv[:] = cmd
                sys.stdout, sys.stderr = io.StringIO(), io.StringIO()
                os.chdir(cwd)
                runpy.run_path(
                    # If passed a Python file, run it directly
                    cli_args[0] if cli_args[0].endswith('.py') else
                    os.path.join(cwd, cli_args[0], cli_args[0] + '.py'),
                    run_name='__main__',
                )
            except SystemExit as e:
                assert e.code in {0, None}
            finally:
                stdout = sys.stdout.getvalue()
                sys.argv[:] = old_sys_argv
                sys.stdout, sys.stderr = old_stdout, old_stderr
                os.chdir(old_cwd)
            for entry in contains:
                assert entry in stdout
            for entry in excludes:
                assert entry not in stdout
            for outpath, options in outputs.items():
                if outpath.startswith('tmp_'):
                    outpath = os.path.join(tmppath, outpath)
                assert os.path.isfile(outpath)
                if 'hash' in options:
                    sha = hashlib.sha256()
                    # For png test files, take a hash of the decoded image
                    # rather than the file.
                    if outpath.endswith('.png'):
                        sha.update(PIL.Image.open(outpath).convert('RGBA').tobytes())  # noqa
                    else:
                        with open(outpath, 'rb') as fptr:
                            while True:
                                data = fptr.read(chunkSize)
                                if not data:
                                    break
                                sha.update(data)
                    hashval = sha.hexdigest()
                    try:
                        assert options['hash'] == hashval
                    except Exception:
                        sys.stderr.write('File hash mismatch: %s\n' % outpath)
                        raise
                if 'contains' in options:
                    data = open(outpath).read(chunkSize)
                    for entry in options['contains']:
                        assert entry in data
        except Exception:
            sys.stderr.write('CMD (cwd %s):\n%r\n' % (cwd, cmd))
            sys.stderr.write('STDOUT:\n%s\n' % stdout.rstrip())
            raise
        finally:
            shutil.rmtree(tmppath)
        return stdout

    def testListCLI(self):
        """
        Test that we can list clis, get help for each, and get the xml spec for
        each.
        """
        from girder.plugins.slicer_cli_web import rest_slicer_cli
        from girder.api.rest import Resource

        restResource = Resource()
        cli_args = (os.environ['CLI_LIST_ENTRYPOINT'], '--list_cli',)
        cli_list = self._runTest(cli_args, contains=['"NucleiDetection"'])
        cli_list = json.loads(cli_list)
        assert 'NucleiDetection' in cli_list
        for cli in cli_list:
            cli_args = (cli, '--help')
            # Each cli's help must mention usage, its own name, and that you
            # can output xml
            self._runTest(cli_args, contains=['usage:', cli, '--xml'])
            cli_args = (cli, '--xml')
            # The xml output needs to have a tile and an executable tag
            xml = self._runTest(cli_args, contains=[
                '<executable>', '<title>', '</executable>', '</title>'])
            if '<' in xml:
                xml = xml[xml.index('<'):]
            try:
                rest_slicer_cli.genHandlerToRunDockerCLI(
                    'dockerimage', os.environ.get('CLI_CWD'), xml, restResource)
            except Exception:
                sys.stderr.write('Failed in generating endpoints for %s' % cli)
                raise

    def testNucleiDetectionDefaults(self):
        cli_args = (
            'NucleiDetection',
            os.path.join(TEST_DATA_DIR, 'TCGA-02-0010-01Z-00-DX4.07de2e55-a8fe-40ee-9e98-bcb78050b9f7-crop.tif'),   # noqa
            'tmp_1.anot',
        )

        # test without ROI
        cli_kwargs = {
            'analysis_roi': '-1.0, -1.0, -1.0, -1.0',
        }
        self._runTest(cli_args, cli_kwargs, outputs={
            'tmp_1.anot': {
                'contains': ['elements'],
                # 'hash': '02b240586412c87ad5cbf349b7c22f80f1df31eef54ed8ee4ad1fd3624a89fa2',  # noqa
            },
        })

        # test with ROI
        cli_kwargs = {
            'analysis_roi': '0.0, 200.0, 512.0, 512.0'
        }
        self._runTest(cli_args, cli_kwargs, outputs={
            'tmp_1.anot': {
                'contains': ['elements'],
                # 'hash': '02b240586412c87ad5cbf349b7c22f80f1df31eef54ed8ee4ad1fd3624a89fa2',  # noqa
            },
        })

    def testColorDeconvolutionDefaults(self):
        self._runTest(
            cli_args=[
                'ColorDeconvolution',
                datastore.fetch('Easy1.png'),
            ] + [f'tmp_out_{i}.png' for i in (1, 2, 3)],
            outputs={
                'tmp_out_1.png': dict(
                    hash='b91d961b7eba8c02a1c067da91fced315fa3922db73f489202a296f4c2304b94',  # noqa
                ),
            },
        )



#!/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import numpy as np
from histomicstk.segmentation.label import trace_object_boundaries
from histomicstk.segmentation.label import delete_border


class TestTraceBoundary:

    def test_trace_boundary(self):

        # test moore neighbor algorithm

        m_neighbor = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
                               [0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
                               [0, 0, 0, 1, 1, 1, 1, 0, 0, 0],
                               [0, 0, 1, 1, 1, 1, 1, 0, 0, 0],
                               [0, 1, 1, 1, 1, 1, 1, 0, 0, 0],
                               [0, 0, 1, 1, 1, 1, 1, 0, 0, 0],
                               [0, 0, 0, 1, 1, 1, 1, 0, 0, 0],
                               [0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
                               [0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
                               [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]], dtype=np.bool)

        # refenece neighbors for isbf
        rx_isbf = [1, 1, 2, 2, 3, 4, 5, 6, 7, 8, 8, 9, 9, 8, 7, 7, 7,
                   7, 6, 6, 5, 5, 5, 4, 4, 3, 3, 3, 3, 2, 1]
        ry_isbf = [7, 8, 8, 7, 6, 6, 6, 6, 6, 7, 8, 8, 7, 7, 6, 5, 4,
                   3, 3, 2, 2, 1, 2, 2, 3, 3, 4, 5, 6, 7, 7]

        # reference neighbors for moore
        rx_moore = [1, 1, 2, 2, 3, 4, 5, 6, 7, 8, 8, 9, 9, 8, 7, 7, 7,
                    7, 6, 5, 4, 3, 3, 3, 3, 2, 1]
        ry_moore = [7, 8, 8, 7, 6, 6, 6, 6, 6, 7, 8, 8, 7, 7, 6, 5, 4,
                    3, 2, 1, 2, 3, 4, 5, 6, 7, 7]

        x_isbf, y_isbf = trace_object_boundaries(
            m_neighbor, simplify_colinear_spurs=False)

        np.testing.assert_allclose(rx_isbf, x_isbf[0])
        np.testing.assert_allclose(ry_isbf, y_isbf[0])

        x_moore, y_moore = trace_object_boundaries(
            m_neighbor, 8, simplify_colinear_spurs=False)

        np.testing.assert_allclose(rx_moore, x_moore[0])
        np.testing.assert_allclose(ry_moore, y_moore[0])


class TestDeleteBorderLabel:

    def test_delete_border(self):

        # check for correct deletion in an image that has border objects
        im_label = np.zeros((10, 10), dtype='int')
        im_label[0:3, 4:7] = 1
        im_label[8:10, 4:7] = 2
        im_label[4:7, 0:3] = 3
        im_label[4:7, 8:10] = 4

        im_label_del = delete_border(im_label)

        np.testing.assert_array_equal(im_label_del, np.zeros_like(im_label))

        # check if it returns original mask if there are no border objects
        im_label = np.zeros((10, 10), dtype='int')
        im_label[4:6, 4:6] = 1

        im_label_del = delete_border(im_label)

        np.testing.assert_array_equal(im_label_del, im_label)



#!/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import numpy as np
import skimage.io
from histomicstk.preprocessing import color_deconvolution as htk_dcv

from .datastore import datastore


class TestMacenko:

    def test_macenko(self):
        im_path = datastore.fetch('Easy1.png')
        im = skimage.io.imread(im_path)[..., :3]

        w = htk_dcv.rgb_separate_stains_macenko_pca(im, 255)

        w_expected = [[0.089411, 0.558021, -0.130574],
                      [0.837138, 0.729935, 0.546981],
                      [0.539635, 0.394725, -0.826899]]

        np.testing.assert_allclose(w, w_expected, atol=1e-6)


class TestColorDeconvolution:

    def test_roundtrip(self):
        im_path = datastore.fetch('Easy1.png')
        im = skimage.io.imread(im_path)[..., :3]

        w = np.array([[0.650, 0.072, 0],
                      [0.704, 0.990, 0],
                      [0.286, 0.105, 0]])

        conv_result = htk_dcv.color_deconvolution(im, w, 255)

        im_reconv = htk_dcv.color_convolution(conv_result.StainsFloat,
                                              conv_result.Wc, 255)

        np.testing.assert_allclose(im, im_reconv, atol=1)



from argparse import Namespace
from datetime import timedelta
from slicer_cli_web import ctk_cli_adjustment  # noqa - imported for side effects
from ctk_cli import CLIArgumentParser
import psutil
import numpy as np
import skimage.measure
import skimage.morphology

import histomicstk.preprocessing.color_deconvolution as htk_cdeconv
import histomicstk.segmentation as htk_seg
import histomicstk.utils as htk_utils

import large_image


# These defaults are only used if girder is not present
# Use memcached by default.
large_image.config.setConfig('cache_backend', 'memcached')
# If memcached is unavilable, specify the fraction of memory that python
# caching is allowed to use.  This is deliberately small.
large_image.config.setConfig('cache_python_memory_portion', 32)


def get_stain_vector(args, index):
    """Get the stain corresponding to args.stain_$index and
    args.stain_$index_vector.  If the former is not "custom", all the
    latter's elements must be -1.

    """
    args = args._asdict() if hasattr(args, '_asdict') else vars(args)
    stain = args['stain_' + str(index)]
    stain_vector = args['stain_' + str(index) + '_vector']
    if all(x == -1 for x in stain_vector):  # Magic default value
        if stain == 'custom':
            raise ValueError('If "custom" is chosen for a stain, '
                             'a stain vector must be provided.')
        return htk_cdeconv.stain_color_map[stain]
    else:
        if stain == 'custom':
            return stain_vector
        raise ValueError('Unless "custom" is chosen for a stain, '
                         'no stain vector may be provided.')


def get_stain_matrix(args, count=3):
    """Get the stain matrix corresponding to the args.stain_$index and
    args.stain_$index_vector arguments for values of index 1 to count.
    Return a numpy array of column vectors.

    """
    return np.array([get_stain_vector(args, i+1) for i in range(count)]).T


def segment_wsi_foreground_at_low_res(ts, lres_size=2048):

    ts_metadata = ts.getMetadata()

    # get image at low-res
    maxSize = max(ts_metadata['sizeX'], ts_metadata['sizeY'])
    maxSize = float(max(maxSize, lres_size))

    downsample_factor = 2.0 ** np.floor(np.log2(maxSize / lres_size))

    fgnd_seg_mag = ts_metadata['magnification'] / downsample_factor

    fgnd_seg_scale = {'magnification': fgnd_seg_mag}

    im_lres, _ = ts.getRegion(
        scale=fgnd_seg_scale,
        format=large_image.tilesource.TILE_FORMAT_NUMPY
    )

    im_lres = im_lres[:, :, :3]

    # compute foreground mask at low-res
    im_fgnd_mask_lres = htk_utils.simple_mask(im_lres)

    return im_fgnd_mask_lres, fgnd_seg_scale


def create_tile_nuclei_bbox_annotations(im_nuclei_seg_mask, tile_info):

    nuclei_annot_list = []

    gx = tile_info['gx']
    gy = tile_info['gy']
    wfrac = tile_info['gwidth'] / np.double(tile_info['width'])
    hfrac = tile_info['gheight'] / np.double(tile_info['height'])

    nuclei_obj_props = skimage.measure.regionprops(im_nuclei_seg_mask)

    for i in range(len(nuclei_obj_props)):
        cx = nuclei_obj_props[i].centroid[1]
        cy = nuclei_obj_props[i].centroid[0]
        width = nuclei_obj_props[i].bbox[3] - nuclei_obj_props[i].bbox[1] + 1
        height = nuclei_obj_props[i].bbox[2] - nuclei_obj_props[i].bbox[0] + 1

        # convert to base pixel coords
        cx = np.round(gx + cx * wfrac, 2)
        cy = np.round(gy + cy * hfrac, 2)
        width = np.round(width * wfrac, 2)
        height = np.round(height * hfrac, 2)

        # create annotation json
        cur_bbox = {
            "type": "rectangle",
            "center": [cx, cy, 0],
            "width": width,
            "height": height,
            "rotation": 0,
            "fillColor": "rgba(0,0,0,0)",
            "lineColor": "rgb(0,255,0)"
        }

        nuclei_annot_list.append(cur_bbox)

    return nuclei_annot_list


def create_tile_nuclei_boundary_annotations(im_nuclei_seg_mask, tile_info):

    nuclei_annot_list = []

    gx = tile_info['gx']
    gy = tile_info['gy']
    wfrac = tile_info['gwidth'] / np.double(tile_info['width'])
    hfrac = tile_info['gheight'] / np.double(tile_info['height'])

    by, bx = htk_seg.label.trace_object_boundaries(im_nuclei_seg_mask,
                                                   trace_all=True)

    for i in range(len(bx)):

        # get boundary points and convert to base pixel space
        num_points = len(bx[i])

        if num_points < 3:
            continue

        cur_points = np.zeros((num_points, 3))
        cur_points[:, 0] = np.round(gx + bx[i] * wfrac, 2)
        cur_points[:, 1] = np.round(gy + by[i] * hfrac, 2)
        cur_points = cur_points.tolist()

        # create annotation json
        cur_annot = {
            "type": "polyline",
            "points": cur_points,
            "closed": True,
            "fillColor": "rgba(0,0,0,0)",
            "lineColor": "rgb(0,255,0)"
        }

        nuclei_annot_list.append(cur_annot)

    return nuclei_annot_list


def create_tile_nuclei_annotations(im_nuclei_seg_mask, tile_info, format):

    if format == 'bbox':

        return create_tile_nuclei_bbox_annotations(im_nuclei_seg_mask,
                                                   tile_info)

    elif format == 'boundary':

        return create_tile_nuclei_boundary_annotations(im_nuclei_seg_mask,
                                                       tile_info)
    else:

        raise ValueError('Invalid value passed for nuclei_annotation_format')


def create_dask_client(args):
    """Create and install a Dask distributed client using args from a
    Namespace, supporting the following attributes:

    - .scheduler: Address of the distributed scheduler, or the
      empty string to start one locally

    """
    import dask
    scheduler = getattr(args, 'scheduler', None)
    num_workers = getattr(args, 'num_workers', 0)
    num_threads_per_worker = getattr(args, 'num_threads_per_worker', 0)

    if scheduler == 'multithreading':
        import dask.threaded
        from multiprocessing.pool import ThreadPool

        if num_threads_per_worker <= 0:
            num_workers = max(1, psutil.cpu_count(logical=False) + num_threads_per_worker)
        else:
            num_workers = num_threads_per_worker
        print('Starting dask thread pool with %d thread(s)' % num_workers)
        dask.config.set(pool=ThreadPool(num_workers))
        dask.config.set(scheduler='threads')
        return

    if scheduler == 'multiprocessing':
        import dask.multiprocessing
        import multiprocessing

        dask.config.set(scheduler='processes')
        if num_workers <= 0:
            num_workers = max(1, psutil.cpu_count(logical=False) + num_workers)

        print('Starting dask multiprocessing pool with %d worker(s)' % num_workers)
        dask.config.set(pool=multiprocessing.Pool(
            num_workers, initializer=dask.multiprocessing.initialize_worker_process))
        return

    import dask.distributed
    if not scheduler:

        if num_workers <= 0:
            num_workers = max(1, psutil.cpu_count(logical=False) + num_workers)
        num_threads_per_worker = (num_threads_per_worker if num_threads_per_worker >= 1 else None)

        print('Creating dask LocalCluster with %d worker(s), %r thread(s) per '
              'worker' % (num_workers, num_threads_per_worker))
        scheduler = dask.distributed.LocalCluster(
            ip='0.0.0.0',  # Allow reaching the diagnostics port externally
            scheduler_port=0,  # Don't expose the scheduler port
            n_workers=num_workers,
            memory_limit=0,
            threads_per_worker=num_threads_per_worker,
            silence_logs=False
        )

    return dask.distributed.Client(scheduler)


def get_region_dict(region, maxRegionSize=None, tilesource=None):
    """Return a dict corresponding to region, checking the region size if
    maxRegionSize is provided.

    The intended use is to be passed via **kwargs, and so either {} is
    returned (for the special region -1,-1,-1,-1) or {'region':
    region_dict}.

    Params
    ------
    region: list
        4 elements -- left, top, width, height -- or all -1, meaning the whole
        slide.
    maxRegionSize: int, optional
        Maximum size permitted of any single dimension
    tilesource: tilesource, optional
        A `large_image` tilesource (or anything with `.sizeX` and `.sizeY`
        properties) that is used to determine the size of the whole slide if
        necessary.  Must be provided if `maxRegionSize` is.

    Returns
    -------
    region_dict: dict
        Either {} (for the special region -1,-1,-1,-1) or
        {'region': region_subdict}

    """

    if len(region) != 4:
        raise ValueError('Exactly four values required for --region')

    useWholeImage = region == [-1] * 4

    if maxRegionSize is not None:
        if tilesource is None:
            raise ValueError('tilesource must be provided if maxRegionSize is')
        if maxRegionSize != -1:
            if useWholeImage:
                size = max(tilesource.sizeX, tilesource.sizeY)
            else:
                size = max(region[-2:])
            if size > maxRegionSize:
                raise ValueError('Requested region is too large!  '
                                 'Please see --maxRegionSize')

    return {} if useWholeImage else dict(
        region=dict(zip(['left', 'top', 'width', 'height'],
                        region)))


def disp_time_hms(seconds):
    """Converts time from seconds to a string of the form hours:minutes:seconds
    """

    return str(timedelta(seconds=seconds))


def splitArgs(args, split='_'):
    """Split a Namespace into a Namespace of Namespaces based on shared
    prefixes.  The string separating the prefix from the rest of the
    argument is determined by the optional "split" parameter.
    Parameters not containing the splitting string are kept as-is.

    """
    def splitKey(k):
        s = k.split(split, 1)
        return (None, s[0]) if len(s) == 1 else s

    args = args._asdict() if hasattr(args, '_asdict') else vars(args)
    firstKeys = {splitKey(k)[0] for k in args}
    result = Namespace()
    for k in firstKeys - {None}:
        setattr(result, k, Namespace())
    for k, v in args.items():
        f, s = splitKey(k)
        if f is None:
            setattr(result, s, v)
        else:
            setattr(getattr(result, f), s, v)
    return result


def sample_pixels(args):
    """Version of histomicstk.utils.sample_pixels that takes a Namespace
    and handles the special default values.

    """
    args = (args._asdict() if hasattr(args, '_asdict') else vars(args)).copy()
    for k in 'magnification', 'sample_fraction', 'sample_approximate_total':
        if args[k] == -1:
            del args[k]
    return htk_utils.sample_pixels(**args)


__all__ = (
    'CLIArgumentParser',
    'create_dask_client',
    'create_tile_nuclei_annotations',
    'create_tile_nuclei_bbox_annotations',
    'create_tile_nuclei_boundary_annotations',
    'disp_time_hms',
    'get_region_dict',
    'get_stain_matrix',
    'get_stain_vector',
    'sample_pixels',
    'segment_wsi_foreground_at_low_res',
    'splitArgs',
)



from histomicstk.cli.utils import CLIArgumentParser
import numpy

from histomicstk.preprocessing.color_deconvolution import rgb_separate_stains_macenko_pca

from histomicstk.cli import utils


def main(args):
    args = utils.splitArgs(args)
    args.macenko.I_0 = numpy.array(args.macenko.I_0)

    utils.create_dask_client(args.dask)
    sample = utils.sample_pixels(args.sample)
    stain_matrix = rgb_separate_stains_macenko_pca(sample.T, **vars(args.macenko))
    with open(args.returnParameterFile, 'w') as f:
        for i, stain in enumerate(stain_matrix.T):
            f.write('stainColor_{} = {}\n'.format(i+1, ','.join(map(str, stain))))


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



import numpy as np
import pandas as pd
import scipy.stats
from skimage.feature import canny
from skimage.measure import regionprops


def compute_gradient_features(im_label, im_intensity,
                              num_hist_bins=10, rprops=None):
    """Calculates gradient features from an intensity image.

    Parameters
    ----------
    im_label : array_like
        A labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.
    im_intensity : array_like
        Intensity image
    num_hist_bins: int, optional
        Number of bins used to computed the gradient histogram of an object.
        Histogram is used to energy and entropy features. Default is 10.
    rprops : output of skimage.measure.regionprops, optional
        rprops = skimage.measure.regionprops( im_label ). If rprops is not
        passed then it will be computed inside which will increase the
        computation time.

    Returns
    -------
    fdata: pandas.DataFrame
        A pandas dataframe containing the gradient features listed below for
        each object/label.

    Notes
    -----
    List of gradient features computed by this function:

    Gradient.Mag.Mean : float
        Mean of gradient data.

    Gradient.Mag.Std : float
        Standard deviation of gradient data.

    Gradient.Mag.Skewness : float
        Skewness of gradient data. Value is 0 when all values are equal.

    Gradient.Mag.Kurtosis : float
        Kurtosis of gradient data. Value is -3 when all values are equal.

    Gradient.Mag.HistEnergy : float
        Energy of the gradient magnitude histogram of object pixels

    Gradient.Mag.HistEnergy : float
        Entropy of the gradient magnitude histogram of object pixels.

    Gradient.Canny.Sum : float
        Sum of canny filtered gradient data.

    Gradient.Canny.Mean : float
        Mean of canny filtered gradient data.

    References
    ----------
    .. [#] Daniel Zwillinger and Stephen Kokoska. "CRC standard probability
       and statistics tables and formulae," Crc Press, 1999.

    """

    # List of feature names
    feature_list = [
        'Gradient.Mag.Mean',
        'Gradient.Mag.Std',
        'Gradient.Mag.Skewness',
        'Gradient.Mag.Kurtosis',
        'Gradient.Mag.HistEntropy',
        'Gradient.Mag.HistEnergy',
        'Gradient.Canny.Sum',
        'Gradient.Canny.Mean',
    ]

    # compute object properties if not provided
    if rprops is None:
        rprops = regionprops(im_label)

    # create pandas data frame containing the features for each object
    numFeatures = len(feature_list)
    numLabels = len(rprops)
    fdata = pd.DataFrame(np.zeros((numLabels, numFeatures)),
                         columns=feature_list)

    Gx, Gy = np.gradient(im_intensity)
    diffG = np.sqrt(Gx**2 + Gy**2)
    cannyG = canny(im_intensity)

    for i in range(numLabels):

        # get gradients of object pixels
        pixelGradients = np.sort(
            diffG[rprops[i].coords[:, 0], rprops[i].coords[:, 1]]
        )

        # compute mean
        fdata.at[i, 'Gradient.Mag.Mean'] = np.mean(pixelGradients)

        # compute standard deviation
        fdata.at[i, 'Gradient.Mag.Std'] = np.std(pixelGradients)

        # compute skewness
        fdata.at[i, 'Gradient.Mag.Skewness'] = scipy.stats.skew(pixelGradients)

        # compute kurtosis
        fdata.at[i, 'Gradient.Mag.Kurtosis'] = \
            scipy.stats.kurtosis(pixelGradients)

        # compute intensity histogram
        hist, bins = np.histogram(pixelGradients, bins=num_hist_bins)
        prob = hist/np.sum(hist, dtype=np.float32)

        # compute entropy
        fdata.at[i, 'Gradient.Mag.HistEntropy'] = scipy.stats.entropy(prob)

        # compute energy
        fdata.at[i, 'Gradient.Mag.HistEnergy'] = np.sum(prob**2)

        bw_canny = cannyG[rprops[i].coords[:, 0], rprops[i].coords[:, 1]]
        canny_sum = np.sum(bw_canny).astype('float')

        fdata.at[i, 'Gradient.Canny.Sum'] = canny_sum

        fdata.at[i, 'Gradient.Canny.Mean'] = canny_sum / len(pixelGradients)

    return fdata



from histomicstk.cli.utils import CLIArgumentParser

from histomicstk.preprocessing.color_normalization import background_intensity

from histomicstk.cli import utils


def main(args):
    other_args = {'returnParameterFile', 'scheduler'}
    kwargs = {k: v for k, v in vars(args).items()
              if k not in other_args}
    # Allow (some) default parameters to work.  Assume certain values
    # are not valid.
    for k in 'sample_fraction', 'sample_approximate_total':
        if kwargs[k] == -1:
            del kwargs[k]

    utils.create_dask_client(args)
    I_0 = background_intensity(**kwargs)
    with open(args.returnParameterFile, 'w') as f:
        f.write('BackgroundIntensity = ' + ','.join(map(str, I_0)) + '\n')


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



from collections import namedtuple

import numpy as np
from numpy import linalg
from pandas import DataFrame
from scipy.spatial import cKDTree as KDTree, Voronoi
from scipy import sparse
from scipy.sparse.csgraph import minimum_spanning_tree
from scipy.spatial.distance import pdist

PopStats = namedtuple('PopStats', ['mean', 'stddev', 'min_max_ratio', 'disorder'])
PolyProps = namedtuple('PolyProps', ['area', 'peri', 'max_dist'])
TriProps = namedtuple('TriProps', ['sides', 'area'])
DensityProps = namedtuple('DensityProps', ['neighbors_in_distance',
                                           'distance_for_neighbors'])
Props = namedtuple('Props', ['voronoi', 'delaunay', 'mst_branches', 'density'])


def compute_global_cell_graph_features(
        centroids,
        neighbor_distances=10. * np.arange(1, 6),
        neighbor_counts=(3, 5, 7),
):
    r"""Compute global (i.e., not per-nucleus) features of the nuclei with
    the given centroids based on the partitioning of the space into
    Voronoi cells and on the induced graph structure.

    Parameters
    ----------
    centroids : array_like
        Nx2 numpy array of nuclear centroids
    neighbor_distances : array_like
        Radii to count neighbors in
    neighbor_counts : sequence
        Sequence of numbers of neighbors, each of which is used to
        compute statistics relating to the distance required to reach
        that many neighbors.

    Returns
    -------
    props : pandas.DataFrame
        A single-row DataFrame with the following columns:

        - voronoi\_...: Voronoi diagram features

          - area\_...: Polygon area features
          - peri\_...: Polygon perimeter features
          - max_dist\_...: Maximum distance in polygon features

        - delaunay\_...: Delaunay triangulation features

          - sides\_...: Triangle side length features
          - area\_...: Triangle area features

        - mst_branches\_...: Minimum spanning tree branch features
        - density\_...: Density features

          - neighbors_in_distance\_...

            - 0, 1, ..., len(neighbor_distances) - 1: Neighbor count
              within given radius features.

          - distance_for_neighbors\_...

            - 0, 1, ..., len(neighbor_counts) - 1: Minimum distance to
              enclose count neighbors features

        The "..."s are meant to signify that what precedes is the
        start of a column name.  At the end of each column name is one
        of 'mean', 'stddev', 'min_max_ratio', and 'disorder'.
        'min_max_ratio' is the minimum-to-maximum ratio, and disorder
        is stddev / (mean + stddev).

    Note
    ----
    The indices for the density features are with respect to the
    *sorted* values of the corresponding argument sequence.

    References
    ----------
    .. [#] Doyle, S., Agner, S., Madabhushi, A., Feldman, M., & Tomaszewski, J.
       (2008, May).  Automated grading of breast cancer histopathology using
       spectral clustering with textural and architectural image features.
       In Biomedical Imaging: From Nano to Macro, 2008.  ISBI 2008.
       5th IEEE International Symposium on (pp. 496-499).  IEEE.

    """
    return _flatten_to_dataframe(_compute_global_cell_graph_features(
        centroids,
        neighbor_distances,
        neighbor_counts,
    ))


def _compute_global_cell_graph_features(
        centroids,
        neighbor_distances,
        neighbor_counts,
):
    """Internal support for compute_global_cell_graph_features that
    returns its result in a nested nametuple structure instead of a
    pandas DataFrame.

    """
    vor = Voronoi(centroids)
    centroids = vor.points
    vertices = vor.vertices

    regions = [r for r in vor.regions if r and -1 not in r]
    areas = np.stack(_poly_area(vertices[r]) for r in regions)
    peris = np.stack(_poly_peri(vertices[r]) for r in regions)
    max_dists = np.stack(pdist(vertices[r]).max() for r in regions)
    poly_props = PolyProps._make(map(_pop_stats, (areas, peris, max_dists)))

    # Assume that each Voronoi vertex is on exactly three ridges.
    ridge_points = vor.ridge_points
    # This isn't exactly the collection of sides, since if they should
    # be counted per-triangle then we weight border ridges wrong
    # relative to ridges that are part of two triangles.
    ridge_lengths = _dist(*np.swapaxes(centroids[ridge_points], 0, 1))
    sides = ridge_lengths
    # Point indices of each triangle
    tris = [[] for _ in vertices]
    for p, ri in enumerate(vor.point_region):
        for vi in vor.regions[ri]:
            if vi != -1:
                tris[vi].append(p)
    # This will only fail in particular symmetrical cases where a
    # Voronoi vertex is associated with more than three centroids.
    # Since this should be unlikely in practice, we don't handle it
    # beyond throwing an AssertionError
    assert all(len(t) == 3 for t in tris)
    tris = np.asarray(tris)
    areas = np.stack(_poly_area(centroids[t]) for t in tris)
    tri_props = TriProps._make(map(_pop_stats, (sides, areas)))

    graph = sparse.coo_matrix((ridge_lengths, np.sort(ridge_points).T),
                              (len(centroids), len(centroids)))
    mst = minimum_spanning_tree(graph)
    # Without looking into exactly how minimum_spanning_tree
    # constructs its output, elimate any explicit zeros to be on the
    # safe side.
    mst_branches = _pop_stats(mst.data[mst.data != 0])

    tree = KDTree(centroids)
    neigbors_in_distance = {
        # Yes, we just throw away the actual points
        r: _pop_stats(np.stack(map(len, tree.query_ball_tree(tree, r))) - 1)
        for r in neighbor_distances
    }
    distance_for_neighbors = dict(zip(
        neighbor_counts,
        map(_pop_stats, tree.query(centroids, [c + 1 for c in neighbor_counts])[0].T),
    ))
    density_props = DensityProps(neigbors_in_distance, distance_for_neighbors)

    return Props(poly_props, tri_props, mst_branches, density_props)


def _poly_area(vertices):
    return abs(_poly_signed_area(vertices))


def _poly_signed_area(vertices):
    return .5 * linalg.det(
        np.stack((vertices, np.roll(vertices, -1, axis=-2)), -1)
    ).sum(-1)


def _poly_peri(vertices):
    return _dist(vertices, np.roll(vertices, -1, axis=-2)).sum(-1)


def _dist(x, y):
    """Compute the distance between two sets of points.  Has signature
    (i),(i)->().

    """
    return (np.subtract(x, y) ** 2).sum(-1) ** .5


def _pop_stats(pop):
    # Filter out outliers (here defined as points more than three
    # standard deviations away from the mean)
    while True:
        mean = pop.mean()
        stddev = pop.std()
        mask = abs(pop - mean) <= 3 * stddev
        if mask.all():
            break
        pop = pop[mask]
    minmaxr = pop.min() / pop.max()
    disorder = stddev / (mean + stddev)
    return PopStats(mean, stddev, minmaxr, disorder)


def _flatten_to_dataframe(nt):
    """Flatten the result of _compute_global_cell_graph_features to the
    DataFrame returned by compute_global_cell_graph_features.

    """
    return DataFrame(_flatten_to_dict(nt), index=[0])


def _flatten_to_dict(nt, prefix=''):
    result = {}
    assert isinstance(nt, (tuple, dict))
    if isinstance(nt, tuple):
        d = nt._asdict()
    else:  # nt is a dict
        # We only have numeric keys, and they may be non-nice floats,
        # so just number them by sort order instead of doing something
        # else for names
        d = {str(i): kv[1] for i, kv in enumerate(sorted(nt.items()))}
    for k, v in d.items():
        if not isinstance(v, (tuple, dict)):
            # Terminate
            result[prefix + k] = v
        else:
            result.update(_flatten_to_dict(v, prefix + k + '_'))
    return result



#!/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import numpy as np
from histomicstk.preprocessing.color_conversion import rgb_to_od
from histomicstk.preprocessing.color_conversion import od_to_rgb
from histomicstk.preprocessing.color_conversion import rgb_to_lab
from histomicstk.preprocessing.color_conversion import lab_to_rgb


class TestColorConversion:

    def test_rgb_to_od(self):

        np.testing.assert_array_almost_equal(
            np.round(rgb_to_od(np.zeros((3, 3, 3)) + 117.0), 4),
            np.zeros((3, 3, 3)) + 35.6158
        )

        # check corner cases
        np.testing.assert_array_almost_equal(
            rgb_to_od(np.zeros((3, 3, 3)) + 255.0),
            np.zeros((3, 3, 3))
        )

        np.testing.assert_array_almost_equal(
            rgb_to_od(np.zeros((3, 3, 3))),
            np.zeros((3, 3, 3)) + 255.0
        )

    def test_od_to_rgb(self):

        np.testing.assert_array_almost_equal(
            od_to_rgb(np.zeros((3, 3, 3)) + 35.6158),
            np.zeros((3, 3, 3)) + 116.99987889
        )

        # check corner cases
        np.testing.assert_array_almost_equal(
            od_to_rgb(np.zeros((3, 3, 3))),
            np.zeros((3, 3, 3)) + 255.0
        )

        np.testing.assert_array_almost_equal(
            od_to_rgb(np.zeros((3, 3, 3)) + 255.0),
            np.zeros((3, 3, 3))
        )

    def test_rgb_to_od_to_rgb(self):

        np.random.seed(1)

        im_rand = np.random.randint(0, 255, (10, 10, 3))

        np.testing.assert_array_almost_equal(
            od_to_rgb(rgb_to_od(im_rand)),
            im_rand
        )

    def test_rgb_to_lab_to_rgb(self):

        np.random.seed(1)

        im_rand = np.random.randint(0, 255, (10, 10, 3))

        np.testing.assert_array_almost_equal(
            np.round(lab_to_rgb(rgb_to_lab(im_rand))),
            im_rand
        )



import numpy as np


def find_stain_index(reference, w):
    """Find the index of the stain column vector in w corresponding to the
    reference vector.  Useful in connection with adaptive
    deconvolution routines in order to find the column corresponding
    with a certain expected stain.

    Parameters
    ----------
    reference : array_like
        1D array that is the stain vector to find
    w : array_like
        2D array of columns the same size as reference.
        The columns should be normalized.

    Returns
    -------
    i : int
        Column of w corresponding to reference

    Notes
    -----
    The index of the vector with the smallest angular distance is returned.

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.separate_stains_macenko_pca
    histomicstk.preprocessing.color_deconvolution.color_deconvolution

    """
    dot_products = np.dot(reference, w)
    return np.argmax(np.abs(dot_products))



import numpy

from histomicstk.cli.utils import CLIArgumentParser

import histomicstk.preprocessing.color_deconvolution as htk_cdeconv

from histomicstk.cli import utils


def main(args):
    args = utils.splitArgs(args)
    args.snmf.I_0 = numpy.array(args.snmf.I_0)

    print(">> Starting Dask cluster and sampling pixels")
    utils.create_dask_client(args.dask)
    sample = utils.sample_pixels(args.sample)

    # Create stain matrix
    print('>> Creating stain matrix')

    args.snmf.w_init = utils.get_stain_matrix(args.stains, 2)

    print(args.snmf.w_init)

    # Perform color deconvolution
    print('>> Performing color deconvolution')

    w_est = htk_cdeconv.rgb_separate_stains_xu_snmf(sample.T, **vars(args.snmf))
    w_est = htk_cdeconv.complement_stain_matrix(w_est)

    with open(args.returnParameterFile, 'w') as f:
        for i, stain in enumerate(w_est.T):
            f.write('stainColor_{} = {}\n'.format(i+1, ','.join(map(str, stain))))


if __name__ == "__main__":
    main(CLIArgumentParser().parse_args())



import os
import pytest
import subprocess
import tempfile
import time
import warnings
import docker
import girder_client


def getTestFilePath(name):
    """
    Return the path to a file in the tests/test_files directory.
    :param name: The name of the file.
    :returns: the path to the file.
    """
    return os.path.join(
        os.path.dirname(os.path.realpath(__file__)), 'test_files', name)


def _get_htk_ipaddr(dclient):
    # search docker containers for a DSA docker container
    # and fetch its IP address
    return list(dclient.containers.list(
        filters={'label': 'HISTOMICSTK_GC_TEST'})[0].attrs[
            'NetworkSettings']['Networks'].values())[0]['IPAddress']


def _connect_girder_client_to_local_dsa(ip):
    # connect a girder client to the local DSA docker
    apiUrl = 'http://%s:8080/api/v1' % ip
    gc = girder_client.GirderClient(apiUrl=apiUrl)
    gc.authenticate('admin', 'password')
    return gc


def _connect_to_existing_local_dsa():
    client = docker.from_env(version='auto')
    ipaddr = _get_htk_ipaddr(client)
    return _connect_girder_client_to_local_dsa(ipaddr)


def _create_and_connect_to_local_dsa():
    # create a local dsa docker and connect to it
    cwd = os.getcwd()
    thisDir = os.path.dirname(os.path.realpath(__file__))
    externdatadir = os.path.join(thisDir, '..', '.tox', 'externaldata')
    if not os.path.exists(externdatadir):
        os.makedirs(externdatadir)
    os.chdir(thisDir)
    outfilePath = os.path.join(
        tempfile.gettempdir(), 'histomicstk_test_girder_log.txt')
    with open(outfilePath, 'w') as outfile:

        # build a DSA docker container locally
        proc = subprocess.Popen([
            'docker-compose', 'up', '--build'],
            close_fds=True, stdout=outfile, stderr=outfile)

        os.chdir(cwd)
        timeout = time.time() + 1200

        # connect to docker and take a look at all its containers
        client = docker.from_env(version='auto')
        while time.time() < timeout:
            try:
                ipaddr = _get_htk_ipaddr(client)
                if ipaddr:
                    gc = _connect_girder_client_to_local_dsa(ipaddr)
                    break
            except Exception:  # noqa
                # warnings.warn(
                #     "Looks like the local DSA docker image is still "
                #     "initializing. Will wait a few seconds and try again.",
                # )
                time.sleep(0.1)

    return gc, proc


# TODO -- refactor to session scope by figuring out pytest issue (bug?)
# See https://docs.pytest.org/en/latest/fixture.html
# Note:
#     We could use scope='session' to create the DSA docker instance once
#     and reuse it for all test modules. However, pytest seems to have a
#     bug when yield is used (as oppoed to return) and it does not run the
#     teardown code properly. Instead, the girderClient fixture is called
#     again between modules, causing a stopIteration error.
#     Until this bug is fixed, we restrict the scope to the "module" level
#     to ensure: 1. Safe teardown, and 2. That edits to the database
#     done by one module do not carry over to the next module.
#

# @pytest.fixture(scope='session')
@pytest.fixture(scope='module')
def girderClient():
    """
    Yield an authenticated girder client that points to the server.

    If a local girder server docker is running, this will connect to it,
    otherwise, this will spin up a local girder server, load it with some
    initial data, and connect to it.

    NOTE: The default behavior initializes the docker image once per module and
    re-uses it for all tests. This means whatever one unit test changes in
    the DSA database is persistent for the next unit test. So if, for example,
    you remove one annotation as part of the first unit test, the next unit
    test will not have access to that annotation. Once all the unit tests are
    done, the database is torn down.

    If, instead, if you would like to run tests *repeatedly* (i.e. prototyping)
    , or you would like the changes written by tests in one module to be
    carried over to the next test module, you may prefer to start the server
    manually. That way you won't worry about unknown wait time till the local
    server is fully initialized. To manually start a DSA docker image, navigate
    to the directory where this file exists, and start the container:
    $ cd HistomicsTK/tests/
    $ docker-compose up --build

    Of course, you need to have docker installed and to either
    run this as sudo or be added to the docker group by the system admins.
    """
    try:
        # First we try to connect to any existing local DSA docker
        yield _connect_to_existing_local_dsa()

    except Exception as e:
        warnings.warn(
            e.__repr__() + "\n"
            "Looks like there's no existing local DSA docker running; "
            "will create one now and try again.",
        )
        # create a local dsa docker and connect to it
        gc, proc = _create_and_connect_to_local_dsa()

        yield gc
        proc.terminate()
        proc.wait()



#! /usr/bin/env python

import os
import sys

from setuptools import find_packages

try:
    from skbuild import setup
except ImportError:
    sys.stderr.write("""scikit-build is required to build from source or run tox.
Please run:
  python -m pip install scikit-build
""")
    # from setuptools import setup
    sys.exit(1)


with open('README.rst') as readme_file:
    readme = readme_file.read()


def prerelease_local_scheme(version):
    """
    Return local scheme version unless building on master in CircleCI.

    This function returns the local scheme version number
    (e.g. 0.0.0.dev<N>+g<HASH>) unless building on CircleCI for a
    pre-release in which case it ignores the hash and produces a
    PEP440 compliant pre-release version number (e.g. 0.0.0.dev<N>).
    """
    from setuptools_scm.version import get_local_node_and_date

    if os.getenv('CIRCLE_BRANCH') in {'master'}:
        return ''
    else:
        return get_local_node_and_date(version)


setup(
    name='histomicstk',
    use_scm_version={'local_scheme': prerelease_local_scheme},
    description='A Python toolkit for Histopathology Image Analysis',
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Kitware, Inc.',
    author_email='developers@digitalslidearchive.net',
    url='https://github.com/DigitalSlideArchive/HistomicsTK',
    packages=find_packages(exclude=['tests', '*_test']),
    package_dir={
        'histomicstk': 'histomicstk',
    },
    include_package_data=True,
    install_requires=[
        'girder_client',
        # scientific packages
        'nimfa>=1.3.2',
        'numpy>=1.12.1',
        'scipy>=0.19.0',
        'Pillow>=3.2.0',
        'pandas>=0.19.2',
        'scikit-image>=0.14.2',
        'scikit-learn>=0.18.1',
        'imageio>=2.3.0',
        'shapely[vectorized]',
        'opencv-python-headless',
        'sqlalchemy',
        'matplotlib',
        'pyvips',
        # dask packages
        'dask[dataframe]>=1.1.0',
        'distributed>=1.21.6',
        # large image; for non-linux systems only install the PIL tile source
        # by default.
        'large-image[sources];sys.platform=="linux"',
        'large-image[sources];sys.platform=="linux2"',
        'large-image[pil];sys.platform!="linux" and sys.platform!="linux2"',
        'girder-slicer-cli-web',
        # cli
        'ctk-cli',
    ],
    license='Apache Software License 2.0',
    keywords='histomicstk',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    zip_safe=False,
    python_requires='>=3.6',
)



import numpy as np
from scipy.ndimage.morphology import distance_transform_edt


def dilate_xor(im_label, neigh_width=8):
    """Computes a label mask highlighting a ring-like neighborhood of each
    object or region in a given label mask

    Parameters
    ----------
    im_label : array_like
        A labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.

    neigh_width : float, optional
        The width of the ring-like neighborhood around each object.

    Returns
    -------
    im_neigh_label : array_like
        A labeled mask image highlighting pixels in a ring-like neighborhood of
        width upto `neigh_width` around each object in the given label mask.
        The intensity of each pixel in the ring-like neighborhood is set
        equal to the label of the closest object in the given label mask.
        other pixels (including the ones inside objects) are set to zero.

    """

    # For each background pixel compute the distance to the nearest object and
    # the indices of the nearest object pixel
    im_dist, closest_obj_ind = distance_transform_edt(im_label == 0,
                                                      return_indices=True)
    closest_obj_rind, closest_obj_cind = closest_obj_ind

    # Get indices of background pixels within a given distance from an object
    neigh_rind, neigh_cind = np.where(
        np.logical_and(im_dist > 0, im_dist <= neigh_width)
    )

    # generate labeled neighborhood mask
    im_neigh_label = np.zeros_like(im_label)

    im_neigh_label[neigh_rind, neigh_cind] = im_label[
        closest_obj_rind[neigh_rind, neigh_cind],
        closest_obj_cind[neigh_rind, neigh_cind]]

    return im_neigh_label



#!/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import histomicstk as htk
import numpy as np
import skimage.feature


class TestGLCMMatrixGeneration:

    def test_graycomatrixext(self):

        image = np.array([[0, 0, 1, 1],
                          [0, 0, 1, 1],
                          [0, 2, 2, 2],
                          [2, 2, 3, 3]], dtype=np.uint8)

        # test 0, 45, 90, 135 degree offsets
        res_skim = skimage.feature.greycomatrix(
            image, [1],
            [0, np.pi/4.0, np.pi/2.0, 3*np.pi/4.0],
            levels=4
        )

        res_htk = htk.features.graycomatrixext(
            image, num_levels=4, gray_limits=[0, 3],
            offsets=np.array([[0, 1], [1, 1], [1, 0], [1, -1]])
        )

        np.testing.assert_allclose(np.squeeze(res_htk), np.squeeze(res_skim))

        # test 0, 45, 90, 135 degree offsets - normalized
        res_skim = skimage.feature.greycomatrix(
            image, [1],
            [0, np.pi/4.0, np.pi/2.0, 3*np.pi/4.0],
            levels=4, normed=True
        )

        res_htk = htk.features.graycomatrixext(
            image, num_levels=4, gray_limits=[0, 3], normed=True,
            offsets=np.array([[0, 1], [1, 1], [1, 0], [1, -1]])
        )

        np.testing.assert_allclose(np.squeeze(res_htk), np.squeeze(res_skim))

        # test 0, 45, 90, 135 degree offsets - symmetric
        res_skim = skimage.feature.greycomatrix(
            image, [1],
            [0, np.pi/4.0, np.pi/2.0, 3*np.pi/4.0],
            levels=4, symmetric=True
        )

        res_htk = htk.features.graycomatrixext(
            image, num_levels=4, gray_limits=[0, 3], symmetric=True,
            offsets=np.array([[0, 1], [1, 1], [1, 0], [1, -1]])
        )

        np.testing.assert_allclose(np.squeeze(res_htk), np.squeeze(res_skim))



import os
import json
import time

import numpy as np
import dask

import histomicstk.preprocessing.color_normalization as htk_cnorm
import histomicstk.preprocessing.color_deconvolution as htk_cdeconv
import histomicstk.segmentation.nuclear as htk_nuclear
import histomicstk.segmentation.label as htk_seg_label
import histomicstk.utils as htk_utils

import large_image

from histomicstk.cli.utils import CLIArgumentParser

from histomicstk.cli import utils as cli_utils

import logging
logging.basicConfig(level=logging.CRITICAL)


def detect_tile_nuclei(slide_path, tile_position, args, it_kwargs,
                       src_mu_lab=None, src_sigma_lab=None):

    # get slide tile source
    ts = large_image.getTileSource(slide_path)

    # get requested tile
    tile_info = ts.getSingleTile(
        tile_position=tile_position,
        format=large_image.tilesource.TILE_FORMAT_NUMPY,
        **it_kwargs)

    # get tile image
    im_tile = tile_info['tile'][:, :, :3]

    # perform color normalization
    im_nmzd = htk_cnorm.reinhard(im_tile,
                                 args.reference_mu_lab,
                                 args.reference_std_lab,
                                 src_mu=src_mu_lab,
                                 src_sigma=src_sigma_lab)

    # perform color decovolution
    w = cli_utils.get_stain_matrix(args)

    im_stains = htk_cdeconv.color_deconvolution(im_nmzd, w).Stains

    im_nuclei_stain = im_stains[:, :, 0].astype(np.float)

    # segment nuclear foreground
    im_nuclei_fgnd_mask = im_nuclei_stain < args.foreground_threshold

    # segment nuclei
    im_nuclei_seg_mask = htk_nuclear.detect_nuclei_kofahi(
        im_nuclei_stain,
        im_nuclei_fgnd_mask,
        args.min_radius,
        args.max_radius,
        args.min_nucleus_area,
        args.local_max_search_radius
    )

    # Delete border nuclei
    if args.ignore_border_nuclei is True:

        im_nuclei_seg_mask = htk_seg_label.delete_border(im_nuclei_seg_mask)

    # generate nuclei annotations
    nuclei_annot_list = []

    flag_nuclei_found = np.any(im_nuclei_seg_mask)

    if flag_nuclei_found:
        nuclei_annot_list = cli_utils.create_tile_nuclei_annotations(
            im_nuclei_seg_mask, tile_info, args.nuclei_annotation_format)

    return nuclei_annot_list


def main(args):

    total_start_time = time.time()

    print('\n>> CLI Parameters ...\n')

    print(args)

    if not os.path.isfile(args.inputImageFile):
        raise OSError('Input image file does not exist.')

    if len(args.reference_mu_lab) != 3:
        raise ValueError('Reference Mean LAB should be a 3 element vector.')

    if len(args.reference_std_lab) != 3:
        raise ValueError('Reference Stddev LAB should be a 3 element vector.')

    if len(args.analysis_roi) != 4:
        raise ValueError('Analysis ROI must be a vector of 4 elements.')

    if np.all(np.array(args.analysis_roi) == -1):
        process_whole_image = True
    else:
        process_whole_image = False

    #
    # Initiate Dask client
    #
    print('\n>> Creating Dask client ...\n')

    start_time = time.time()

    c = cli_utils.create_dask_client(args)

    print(c)

    dask_setup_time = time.time() - start_time
    print('Dask setup time = {}'.format(
        cli_utils.disp_time_hms(dask_setup_time)))

    #
    # Read Input Image
    #
    print('\n>> Reading input image ... \n')

    ts = large_image.getTileSource(args.inputImageFile)

    ts_metadata = ts.getMetadata()

    print(json.dumps(ts_metadata, indent=2))

    is_wsi = ts_metadata['magnification'] is not None

    #
    # Compute tissue/foreground mask at low-res for whole slide images
    #
    if is_wsi and process_whole_image:

        print('\n>> Computing tissue/foreground mask at low-res ...\n')

        start_time = time.time()

        im_fgnd_mask_lres, fgnd_seg_scale = \
            cli_utils.segment_wsi_foreground_at_low_res(ts)

        fgnd_time = time.time() - start_time

        print('low-res foreground mask computation time = {}'.format(
            cli_utils.disp_time_hms(fgnd_time)))

    #
    # Compute foreground fraction of tiles in parallel using Dask
    #
    tile_fgnd_frac_list = [1.0]

    it_kwargs = {
        'tile_size': {'width': args.analysis_tile_size},
        'scale': {'magnification': args.analysis_mag},
    }

    if not process_whole_image:

        it_kwargs['region'] = {
            'left':   args.analysis_roi[0],
            'top':    args.analysis_roi[1],
            'width':  args.analysis_roi[2],
            'height': args.analysis_roi[3],
            'units':  'base_pixels'
        }

    if is_wsi:

        print('\n>> Computing foreground fraction of all tiles ...\n')

        start_time = time.time()

        num_tiles = ts.getSingleTile(**it_kwargs)['iterator_range']['position']

        print(f'Number of tiles = {num_tiles}')

        if process_whole_image:

            tile_fgnd_frac_list = htk_utils.compute_tile_foreground_fraction(
                args.inputImageFile, im_fgnd_mask_lres, fgnd_seg_scale,
                it_kwargs
            )

        else:

            tile_fgnd_frac_list = np.full(num_tiles, 1.0)

        num_fgnd_tiles = np.count_nonzero(
            tile_fgnd_frac_list >= args.min_fgnd_frac)

        percent_fgnd_tiles = 100.0 * num_fgnd_tiles / num_tiles

        fgnd_frac_comp_time = time.time() - start_time

        print('Number of foreground tiles = {:d} ({:2f}%%)'.format(
            num_fgnd_tiles, percent_fgnd_tiles))

        print('Tile foreground fraction computation time = {}'.format(
            cli_utils.disp_time_hms(fgnd_frac_comp_time)))

    #
    # Compute reinhard stats for color normalization
    #
    src_mu_lab = None
    src_sigma_lab = None

    if is_wsi and process_whole_image:

        print('\n>> Computing reinhard color normalization stats ...\n')

        start_time = time.time()

        src_mu_lab, src_sigma_lab = htk_cnorm.reinhard_stats(
            args.inputImageFile, 0.01, magnification=args.analysis_mag)

        rstats_time = time.time() - start_time

        print('Reinhard stats computation time = {}'.format(
            cli_utils.disp_time_hms(rstats_time)))

    #
    # Detect nuclei in parallel using Dask
    #
    print('\n>> Detecting nuclei ...\n')

    start_time = time.time()

    tile_nuclei_list = []

    for tile in ts.tileIterator(**it_kwargs):

        tile_position = tile['tile_position']['position']

        if is_wsi and tile_fgnd_frac_list[tile_position] <= args.min_fgnd_frac:
            continue

        # detect nuclei
        cur_nuclei_list = dask.delayed(detect_tile_nuclei)(
            args.inputImageFile,
            tile_position,
            args, it_kwargs,
            src_mu_lab, src_sigma_lab
        )

        # append result to list
        tile_nuclei_list.append(cur_nuclei_list)

    tile_nuclei_list = dask.delayed(tile_nuclei_list).compute()

    nuclei_list = [anot
                   for anot_list in tile_nuclei_list for anot in anot_list]

    nuclei_detection_time = time.time() - start_time

    print('Number of nuclei = {}'.format(len(nuclei_list)))

    print('Nuclei detection time = {}'.format(
        cli_utils.disp_time_hms(nuclei_detection_time)))

    #
    # Write annotation file
    #
    print('\n>> Writing annotation file ...\n')

    annot_fname = os.path.splitext(
        os.path.basename(args.outputNucleiAnnotationFile))[0]

    annotation = {
        "name":     annot_fname + '-nuclei-' + args.nuclei_annotation_format,
        "elements": nuclei_list
    }

    with open(args.outputNucleiAnnotationFile, 'w') as annotation_file:
        json.dump(annotation, annotation_file, indent=2, sort_keys=False)

    total_time_taken = time.time() - total_start_time

    print('Total analysis time = {}'.format(
        cli_utils.disp_time_hms(total_time_taken)))


if __name__ == "__main__":

    main(CLIArgumentParser().parse_args())



#!/usr/bin/env python3
"""
Created on Sun Oct 20 00:14:03 2019.

@author: mtageld
"""
import pytest
import numpy as np
from skimage.transform import resize
from histomicstk.saliency.tissue_detection import get_tissue_mask
from histomicstk.annotations_and_masks.annotation_and_mask_utils import (
    get_image_from_htk_response)
from histomicstk.preprocessing.color_conversion import lab_mean_std
from histomicstk.preprocessing.color_normalization import reinhard
from histomicstk.preprocessing.augmentation.\
    color_augmentation import rgb_perturb_stain_concentration
from histomicstk.preprocessing.color_normalization.\
    deconvolution_based_normalization import deconvolution_based_normalization
import os
import sys
thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../tests'))
# import htk_test_utilities as utilities  # noqa
from htk_test_utilities import girderClient, getTestFilePath  # noqa
# # for protyping
# from tests.htk_test_utilities import _connect_to_existing_local_dsa
# girderClient = _connect_to_existing_local_dsa()


class Cfg:
    def __init__(self):
        self.gc = None
        self.tissue_rgb = None
        self.mask_out = None


cfg = Cfg()


@pytest.mark.usefixtures('girderClient')  # noqa
def test_prep(girderClient):  # noqa

    cfg.gc = girderClient

    iteminfo = cfg.gc.get('/item', parameters={
        'text': "TCGA-A2-A0YE-01Z-00-DX1"})[0]

    # get RGB region at a small magnification
    MAG = 1.5
    getStr = "/item/%s/tiles/region?left=%d&right=%d&top=%d&bottom=%d&encoding=PNG" % (
        iteminfo['_id'], 46890, 50000, 40350, 43000
        ) + "&magnification=%.2f" % MAG
    cfg.tissue_rgb = get_image_from_htk_response(
        cfg.gc.get(getStr, jsonResp=False))

    # get mask of things to ignore
    cfg.mask_out, _ = get_tissue_mask(
        cfg.tissue_rgb, deconvolve_first=False,
        n_thresholding_steps=1, sigma=1.5, min_size=30)
    cfg.mask_out = resize(
        cfg.mask_out == 0, output_shape=cfg.tissue_rgb.shape[:2],
        order=0, preserve_range=True) == 1


class TestColorNormalization():
    """Test color normalization."""

    def test_reinhard(self):
        """Test reinhard."""
        # # SANITY CHECK! normalize to LAB mean and std from SAME slide
        # mean_lab, std_lab = lab_mean_std(tissue_rgb)
        # tissue_rgb_normalized = reinhard(
        #     tissue_rgb, target_mu=mean_lab, target_sigma=std_lab)
        # # we expect the images to be (almost) exactly the same
        # assert np.mean(tissue_rgb - tissue_rgb_normalized) < 1

        # color norm. standard (from TCGA-A2-A3XS-DX1, Amgad et al, 2019)
        cnorm = {
            'mu': np.array([8.74108109, -0.12440419, 0.0444982]),
            'sigma': np.array([0.6135447, 0.10989545, 0.0286032]),
        }

        # Normalize to pre-set color standard (unmasked)
        tissue_rgb_normalized = reinhard(
            cfg.tissue_rgb, target_mu=cnorm['mu'], target_sigma=cnorm['sigma'])

        # check that it matches
        mean_lab, std_lab = lab_mean_std(tissue_rgb_normalized)
        assert all(np.abs(mean_lab - cnorm['mu']) < [0.1, 0.1, 0.1])
        assert all(np.abs(std_lab - cnorm['sigma']) < [0.1, 0.1, 0.1])

        # Do MASKED normalization to preset standard
        tissue_rgb_normalized = reinhard(
            cfg.tissue_rgb, target_mu=cnorm['mu'], target_sigma=cnorm['sigma'],
            mask_out=cfg.mask_out)

        # check that it matches
        mean_lab, std_lab = lab_mean_std(
            tissue_rgb_normalized, mask_out=cfg.mask_out)
        assert all(np.abs(mean_lab - cnorm['mu']) < [0.1, 0.1, 0.1])
        assert all(np.abs(std_lab - cnorm['sigma']) < [0.1, 0.1, 0.1])

    def test_macenko_normalization(self):
        """Test macenko_pca normalization."""
        stain_unmixing_routine_params = {
            'stains': ['hematoxylin', 'eosin'],
            'stain_unmixing_method': 'macenko_pca',
        }

        # TCGA-A2-A3XS-DX1_xmin21421_ymin37486_.png, Amgad et al, 2019)
        # for macenco (obtained using rgb_separate_stains_macenko_pca()
        # and using reordered such that columns are the order:
        # Hamtoxylin, Eosin, Null
        W_target = np.array([
            [0.5807549, 0.08314027, 0.08213795],
            [0.71681094, 0.90081588, 0.41999816],
            [0.38588316, 0.42616716, -0.90380025]
        ])

        # Macenko - Unmasked, using default, 'idealized' W_target"
        tissue_rgb_normalized = deconvolution_based_normalization(
            cfg.tissue_rgb,
            stain_unmixing_routine_params=stain_unmixing_routine_params)
        assert tuple(
            [int(tissue_rgb_normalized[..., i].mean()) for i in range(3)]
            ) == (183, 121, 212)

        # Macenko - Unmasked, using W_target from good image
        tissue_rgb_normalized = deconvolution_based_normalization(
            cfg.tissue_rgb, W_target=W_target,
            stain_unmixing_routine_params=stain_unmixing_routine_params)
        assert tuple(
            [int(tissue_rgb_normalized[..., i].mean()) for i in range(3)]
            ) == (188, 125, 175)

        # Macenko - Masked, using W_target from good image
        tissue_rgb_normalized = deconvolution_based_normalization(
            cfg.tissue_rgb, W_target=W_target, mask_out=cfg.mask_out,
            stain_unmixing_routine_params=stain_unmixing_routine_params)
        assert tuple(
            [int(tissue_rgb_normalized[..., i].mean()) for i in range(3)]
            ) == (188, 125, 175)


class TestColorAugmentation:
    """Test color augmentation."""

    def test_rgb_perturb_stain_concentration(self):
        """Test rgb_perturb_stain_concentration."""
        # for reproducibility
        np.random.seed(0)

        # Unmasked
        augmented_rgb = rgb_perturb_stain_concentration(cfg.tissue_rgb)
        assert tuple(
            [int(augmented_rgb[..., i].mean()) for i in range(3)]
            ) == (178, 115, 154)

        # Masked
        augmented_rgb = rgb_perturb_stain_concentration(
            cfg.tissue_rgb, mask_out=cfg.mask_out)
        assert tuple(
            [int(augmented_rgb[..., i].mean()) for i in range(3)]
            ) == (174, 101, 139)



#!/usr/bin/env python

import os
import subprocess

script = """
python --version && \\
pip install --upgrade pip && \\
pip install 'large_image[tiff,openslide,pil]' \\
  -f https://girder.github.io/large_image_wheels && \\
pip install histomicstk -f /wheels && \\
echo 'Test basic import of histomicstk' && \\
python -c 'import histomicstk' && \\
curl https://data.kitware.com/api/v1/file/5899dd6d8d777f07219fcb23/download -LJ -o tcga.svs && \\
export CLIPATH=`python -c 'import histomicstk,os;print(os.path.dirname( \\
  histomicstk.__file__))'`/cli && \\
python "$CLIPATH/NucleiDetection/NucleiDetection.py" tcga.svs sample.anot \\
  --analysis_roi="7000,7000,3000,3000" && \\
true"""

containers = [
    'python:3.6',
    'python:3.7',
    'python:3.8',
    'python:3.9',
    'centos/python-36-centos7',
]

for container in containers:
    print('---- Testing in %s ----' % container)
    subprocess.check_call([
        'docker', 'run', '-v',
        '%s/wheels:/wheels' % os.path.dirname(os.path.realpath(__file__)),
        '--rm', container, 'sh', '-c', script])

# To test manually, run a container such as
#  docker run -v `pwd`/docs:/wheels --rm -it python:3.7 bash
# and then enter the script commands directly



import os
import json
import colorsys

import numpy as np
import pandas as pd
try:
    import joblib
except ImportError:
    # Versions of scikit-learn before 0.21 had joblib internally
    from sklearn.externals import joblib
import dask.dataframe as dd

from histomicstk.cli.utils import CLIArgumentParser

from histomicstk.cli import utils as cli_utils

import logging
logging.basicConfig(level=logging.CRITICAL)


def gen_distinct_rgb_colors(n, seed=None):
    """
    Generates N visually distinct RGB colors

    Parameters
    ----------
    n : int
        Number of distinct RGB colors to output

    seed: int or array_like, optional
        Seed for the random number generator. See documentation of
        `numpy.random.seed` for more details

    Returns
    -------
    colors_list : list
        A list of n RGB colors

    References
    ----------
    .. [#] http://martin.ankerl.com/2009/12/09/how-to-create-random-colors-programmatically/  # noqa

    """

    np.random.seed(seed)
    h = np.random.random()
    np.random.seed(None)

    golden_ratio_conjugate = (np.sqrt(5) - 1) / 2.0

    color_list = [colorsys.hsv_to_rgb((h + i * golden_ratio_conjugate) % 1,
                                      1.0, 1.0)
                  for i in range(n)]

    return color_list


def read_feature_file(args):

    fname, feature_file_format = os.path.splitext(args.inputNucleiFeatureFile)

    if feature_file_format == '.csv':

        ddf = dd.read_csv(args.inputNucleiFeatureFile)

    elif feature_file_format == '.h5':

        ddf = dd.read_hdf(args.inputNucleiFeatureFile, 'Features')

    else:
        raise ValueError('Extension of output feature file must be .csv or .h5')

    return ddf


def check_args(args):

    if not os.path.isfile(args.inputImageFile):
        raise OSError('Input image file does not exist.')

    if not os.path.isfile(args.inputModelFile):
        raise OSError('Input model file does not exist.')


def main(args):

    print('\n>> CLI Parameters ...\n')

    print(args)

    #
    # Initiate Dask client
    #
    print('\n>> Creating Dask client ...\n')

    c = cli_utils.create_dask_client(args)

    print(c)

    #
    # read model file
    #
    print('\n>> Loading classification model ...\n')

    clf_model = joblib.load(args.inputModelFile)

    #
    # read feature file
    #
    print('\n>> Loading nuclei feature file ...\n')

    ddf = read_feature_file(args)

    if len(ddf.columns) != clf_model.n_features_:

        raise ValueError('The number of features of the classification model '
                         'and the input feature file do not match.')

    #
    # read nuclei annotation file
    #
    print('\n>> Loading nuclei annotation file ...\n')

    with open(args.inputNucleiAnnotationFile) as f:

        nuclei_annot_list = json.load(f)['elements']

    if len(nuclei_annot_list) != len(ddf.index):

        raise ValueError('The number of nuclei in the feature file and the '
                         'annotation file do not match')

    #
    # Perform nuclei classification
    #
    print('\n>> Performing nuclei classification using Dask ...\n')

    def predict_nuclei_class_prob(df, clf_model):

        return pd.DataFrame(data=clf_model.predict_proba(df.as_matrix()),
                            columns=clf_model.classes_)

    outfmt = pd.DataFrame(columns=clf_model.classes_, dtype=np.float64)

    df_class_prob = ddf.map_partitions(predict_nuclei_class_prob, clf_model,
                                       meta=outfmt).compute()

    pred_class = df_class_prob.idxmax(axis=1)

    #
    # Group nuclei annotations by class
    #
    print('\n>> Grouping nuclei annotations by class ...\n')

    num_classes = len(clf_model.classes_)

    nuclei_annot_by_class = {c: [] for c in clf_model.classes_}

    class_color_map = dict(zip(clf_model.classes_,
                               gen_distinct_rgb_colors(num_classes, seed=1)))

    for i in range(len(nuclei_annot_list)):

        cur_class = pred_class.iloc[i]

        cur_anot = nuclei_annot_list[i]
        cur_anot['lineColor'] = 'rgb(%s)' % ','.join(
            [str(int(round(col*255))) for col in class_color_map[cur_class]])
        nuclei_annot_by_class[cur_class].append(cur_anot)

    #
    # Write annotation file
    #
    print('\n>> Writing classified nuclei annotation file ...\n')

    annot_fname = os.path.splitext(
        os.path.basename(args.outputNucleiAnnotationFile))[0]

    annotation = []
    for c in clf_model.classes_:

        annotation.append({
            "name": annot_fname + '-nuclei-class-' + str(c),
            "elements": nuclei_annot_by_class[c]
        })

    with open(args.outputNucleiAnnotationFile, 'w') as annotation_file:
        json.dump(annotation, annotation_file, indent=2, sort_keys=False)


if __name__ == "__main__":
    main(CLIArgumentParser().parse_args())



#!/usr/bin/env python

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import os
import numpy as np
from histomicstk.filters.shape import clog, cdog
from skimage.feature import peak_local_max

from .datastore import datastore


def assert_array_almost_equal_neighborhood_lines(im, gt, decimal=4):
    """Wrapper around assert_array_almost_equal to work around scikit-image bug.

    `skimage.transform.resize()` can return different results on different platforms. This
    is an uncorrected bug mostly due to some floating point computation differences on
    different platforms, hardware, or numerical computation library backend [1].

    Due to that bug, the image (array) returned by the `resize()` function may have a few
    lines in the result that differ from the ground truth computed on a different computer.

    The work around implemented in this function compares each line of the input image with
    each line of a ground truth image, but also compares the line before and the line after.
    Only the minimum error is kept. This allows to ignore 'rogue' lines in the image, that
    are different only due to that scikit-image bug.

    [1] https://github.com/scikit-image/scikit-image/issues/3445
    """
    shift = [-1, 0, 1]
    min_array = np.full_like(im, np.inf)
    for s in shift:
        rolled_image = np.roll(im, s, 1)
        for s in shift:
            rolled_gt = np.roll(gt, s, 1)
            abs_array = np.abs(rolled_gt - rolled_image)
            min_array = np.minimum(min_array, abs_array)
    np.testing.assert_array_almost_equal(min_array, np.zeros_like(im), decimal=decimal)


def _sort_list(input_list):
    return input_list[input_list[:, 0].argsort()]


def compare_maxima(input_im, gtruth_im, min_distance=10, threshold_abs=20):
    """Compares image maxima

    Compare that the maxima found in an image matches the maxima found in a ground truth image.
    This function is a wrapper around `skimage.feature.peak_local_max()`. It calls this function
    on both images that are passed as arguments, and asserts if the resulting maxima arrays
    returned by this function match.
    """
    gtruth_coordinates = _sort_list(peak_local_max(gtruth_im, min_distance=min_distance,
                                                   threshold_abs=threshold_abs))
    input_coordinates = _sort_list(peak_local_max(input_im, min_distance=min_distance,
                                                  threshold_abs=threshold_abs))
    np.testing.assert_array_equal(gtruth_coordinates, input_coordinates)


class TestBlobDetectionFilters:

    def test_clog(self):

        im_nuclei_stain_data = np.load(
            datastore.fetch('Easy1_nuclei_stain.npz'))
        im_nuclei_stain = im_nuclei_stain_data['Easy1_nuclei_stain']

        im_nuclei_fgnd_mask_data = np.load(
            datastore.fetch('Easy1_nuclei_fgnd_mask.npz'))
        im_nuclei_fgnd_mask = im_nuclei_fgnd_mask_data['Easy1_nuclei_fgnd_mask']

        sigma_min = 10.0 / np.sqrt(2.0)
        sigma_max = 40.0 / np.sqrt(2.0)

        im_log_max, im_sigma_max = clog(
            im_input=im_nuclei_stain, im_mask=im_nuclei_fgnd_mask,
            sigma_min=sigma_min, sigma_max=sigma_max)

        im_log_max_gtruth_data = np.load(os.path.join(datastore.fetch(
            'Easy1_clog_max.npz')))
        im_log_max_gtruth = im_log_max_gtruth_data['Easy1_clog_max']
        assert_array_almost_equal_neighborhood_lines(im_log_max, im_log_max_gtruth, decimal=4)
        compare_maxima(im_log_max, im_log_max_gtruth)

        im_sigma_max_gtruth_data = np.load(os.path.join(datastore.fetch(
            'Easy1_clog_sigma_max.npz')))
        im_sigma_max_gtruth = im_sigma_max_gtruth_data['Easy1_clog_sigma_max']
        assert_array_almost_equal_neighborhood_lines(im_sigma_max, im_sigma_max_gtruth, decimal=4)

    def test_cdog(self):

        im_nuclei_stain_data = np.load(
            os.path.join(datastore.fetch('Easy1_nuclei_stain.npz')))
        im_nuclei_stain = im_nuclei_stain_data['Easy1_nuclei_stain']

        im_nuclei_fgnd_mask_data = np.load(
            os.path.join(datastore.fetch('Easy1_nuclei_fgnd_mask.npz')))
        im_nuclei_fgnd_mask = im_nuclei_fgnd_mask_data['Easy1_nuclei_fgnd_mask']

        sigma_min = 10.0 / np.sqrt(2.0)
        sigma_max = 40.0 / np.sqrt(2.0)

        im_dog_max, im_sigma_max = cdog(im_nuclei_stain,
                                        im_nuclei_fgnd_mask,
                                        sigma_min, sigma_max)

        im_dog_max_gtruth_data = np.load(
            os.path.join(datastore.fetch('Easy1_cdog_max.npz')))
        im_dog_max_gtruth = im_dog_max_gtruth_data['Easy1_cdog_max']
        assert_array_almost_equal_neighborhood_lines(im_dog_max, im_dog_max_gtruth, decimal=4)
        compare_maxima(im_dog_max, im_dog_max_gtruth)

        im_sigma_max_gtruth_data = np.load(
            os.path.join(datastore.fetch('Easy1_cdog_sigma_max.npz')))
        im_sigma_max_gtruth = im_sigma_max_gtruth_data['Easy1_cdog_sigma_max']
        assert_array_almost_equal_neighborhood_lines(im_sigma_max, im_sigma_max_gtruth, decimal=4)
